# Copyright (C) 2023 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from datetime import datetime

from pontos.version.errors import VersionError
from pontos.version.schemes._pep440 import PEP440Version
from pontos.version.schemes._semantic import SemanticVersion as Version
from pontos.version.schemes._semantic import (
    SemanticVersionCalculator as VersionCalculator,
)


class SemanticVersionTestCase(unittest.TestCase):
    def test_parse_version(self):
        versions = [
            "0.0.1",
            "1.2.3",
            "1.2.3-foo1",
            "1.2.3-a1",
            "1.2.3-alpha1",
            "1.2.3-alpha1-dev1",
            "1.2.3-b1",
            "1.2.3-beta1",
            "1.2.3-beta1-dev1",
            "1.2.3-rc1",
            "1.2.3-rc1-dev1",
            "1.2.3-dev1",
            "1.2.3+foo1",
            "22.4.1",
            "22.4.1-dev1",
            "22.4.1-dev3",
        ]
        for version in versions:
            self.assertEqual(Version.from_string(version), Version(version))

    def test_parsed_version(self):
        versions = [
            "0.0.1",
            "1.2.3",
            "1.2.3-foo1",
            "1.2.3-a1",
            "1.2.3-alpha1",
            "1.2.3-alpha1-dev1",
            "1.2.3-b1",
            "1.2.3-beta1",
            "1.2.3-beta1-dev1",
            "1.2.3-rc1",
            "1.2.3-rc1-dev1",
            "1.2.3-dev1",
            "1.2.3+foo1",
            "22.4.1",
            "22.4.1-dev1",
            "22.4.1-dev3",
        ]
        for version in versions:
            self.assertEqual(
                Version.from_string(version).parsed_version, version
            )

        pep440_version = PEP440Version.from_string("22.4.1.dev1")
        semver_version = Version.from_version(pep440_version)

        self.assertEqual(str(semver_version), "22.4.1-dev1")
        self.assertEqual(semver_version.parsed_version, "22.4.1.dev1")

    def test_parse_error(self):
        versions = [
            "abc",
            "1.2.3d",
            "1.2.3.post1",
            "1.2.3a1",
            "1.2.3b1",
            "1.2.3rc1",
            "1.2.3a1+dev1",
            "22.4.1.dev1",
            "22.4.1.dev3",
        ]

        for version in versions:
            with self.assertRaisesRegex(
                VersionError, "^.* is not valid SemVer string$"
            ):
                Version.from_string(version)

    def test_parse_prerelease_error(self):
        versions = [
            "1.2.3-pos1t1",
            "1.2.3-dev1-dev1",
        ]

        for version in versions:
            with self.assertRaisesRegex(
                VersionError, f"^Invalid prerelease .* in {version}"
            ):
                Version.from_string(version)

    def test_equal(self):
        versions = [
            ("1.0.0", "1.0.0"),
            ("1.0.0+dev1", "1.0.0+dev1"),
            ("1.0.0-dev1", "1.0.0-dev1"),
            ("1.0.0-alpha1", "1.0.0-alpha1"),
            ("1.0.0-alpha1+dev1", "1.0.0-alpha1+dev1"),
            ("1.0.0-alpha1-dev1", "1.0.0-alpha1-dev1"),
            ("1.0.0-beta1", "1.0.0-beta1"),
            ("1.0.0-beta1-dev1", "1.0.0-beta1-dev1"),
            ("1.0.0-rc1", "1.0.0-rc1"),
            ("1.0.0-rc1-dev1", "1.0.0-rc1-dev1"),
        ]
        for version1, version2 in versions:
            self.assertTrue(
                Version.from_string(version1) == Version.from_string(version2),
                f"{version1} does not equal {version2}",
            )

        versions = [
            ("1.0.0", "1.0.1"),
            ("1.0.0", "1.0.0+dev1"),
            ("1.0.0", "1.0.0-alpha1"),
            ("1.0.0", "1.0.0-alpha1"),
            ("1.0.0", "1.0.0-beta1"),
            ("1.0.0+dev1", "1.0.0-dev1"),
            ("1.0.0+dev1", "1.0.0+dev2"),
            ("1.0.0-alpha1", "1.0.0-beta1"),
            ("1.0.0-alpha1", "1.0.0-alpha1+dev1"),
            ("1.0.0-alpha1+dev1", "1.0.0-alpha1+dev2"),
            ("1.0.0-beta1", "1.0.0-beta1+dev1"),
            ("1.0.0-beta1+dev1", "1.0.0-beta1+dev2"),
            ("1.0.0-rc1", "1.0.0-rc1+dev1"),
            ("1.0.0-rc1+dev1", "1.0.0-rc1+dev2"),
        ]
        for version1, version2 in versions:
            self.assertFalse(
                Version.from_string(version1) == Version.from_string(version2),
                f"{version1} equals {version2}",
            )

        other = None
        self.assertFalse(Version.from_string("1.0.0") == other)

        versions = [
            ("1.0.0", object()),
            ("1.0.0", 1),
            ("1.0.0", True),
        ]
        for version1, version2 in versions:
            with self.assertRaisesRegex(ValueError, "Can't compare"):
                self.assertFalse(Version.from_string(version1) == version2)

    def test_equal_pep440_version(self):
        versions = [
            ("1.0.0", "1.0.0"),
            ("1.0.0-dev1", "1.0.0-dev1"),
            ("1.0.0-dev1", "1.0.0.dev1"),
            ("1.0.0-alpha1", "1.0.0-alpha1"),
            ("1.0.0-alpha1", "1.0.0a1"),
            ("1.0.0-alpha1-dev1", "1.0.0-alpha1-dev1"),
            ("1.0.0-alpha1-dev1", "1.0.0a1.dev1"),
            ("1.0.0-beta1", "1.0.0-beta1"),
            ("1.0.0-beta1", "1.0.0b1"),
            ("1.0.0-beta1-dev1", "1.0.0-beta1-dev1"),
            ("1.0.0-beta1-dev1", "1.0.0b1.dev1"),
            ("1.0.0-rc1", "1.0.0-rc1"),
            ("1.0.0-rc1-dev1", "1.0.0-rc1.dev1"),
            ("1.0.0-rc1-dev1", "1.0.0rc1.dev1"),
        ]
        for version1, version2 in versions:
            semver = Version.from_string(version1)
            pep440 = PEP440Version.from_string(version2)
            self.assertTrue(
                semver == pep440,
                f"{semver!r} {version1} does not equal {pep440!r} {version2}",
            )

    def test_not_equal(self):
        versions = [
            ("1.0.0", "1.0.0"),
            ("1.0.0+dev1", "1.0.0+dev1"),
            ("1.0.0-alpha1", "1.0.0-alpha1"),
            ("1.0.0-alpha1+dev1", "1.0.0-alpha1+dev1"),
            ("1.0.0-beta1", "1.0.0-beta1"),
            ("1.0.0-rc1", "1.0.0-rc1"),
        ]
        for version1, version2 in versions:
            self.assertFalse(
                Version.from_string(version1) != Version.from_string(version2),
                f"{version1} does not equal {version2}",
            )

        versions = [
            ("1.0.0", "1.0.1"),
            ("1.0.0", "1.0.0+dev1"),
            ("1.0.0", "1.0.0-alpha1"),
            ("1.0.0", "1.0.0-alpha1"),
            ("1.0.0", "1.0.0-beta1"),
            ("1.0.0+dev1", "1.0.0+dev2"),
            ("1.0.0-alpha1", "1.0.0-beta1"),
            ("1.0.0-alpha1", "1.0.0-alpha1+dev1"),
            ("1.0.0-alpha1+dev1", "1.0.0-alpha1+dev2"),
            ("1.0.0-beta1", "1.0.0-beta1+dev1"),
            ("1.0.0-beta1+dev1", "1.0.0-beta1+dev2"),
            ("1.0.0-rc1", "1.0.0-rc1+dev1"),
            ("1.0.0-rc1+dev1", "1.0.0-rc1+dev2"),
        ]
        for version1, version2 in versions:
            self.assertTrue(
                Version.from_string(version1) != Version.from_string(version2),
                f"{version1} equals {version2}",
            )

        versions = [
            ("1.0.0", "abc"),
            ("1.0.0", None),
        ]
        for version1, version2 in versions:
            self.assertTrue(Version.from_string(version1) != version2)

        versions = [
            ("1.0.0", object()),
            ("1.0.0", 1),
            ("1.0.0", True),
        ]
        for version1, version2 in versions:
            with self.assertRaisesRegex(ValueError, "Can't compare"):
                self.assertFalse(Version.from_string(version1) != version2)

    def test_greater_then(self):
        versions = [
            ("1.0.0", "0.9.9999"),
            ("1.0.1", "1.0.0"),
            ("1.0.0", "1.0.0-dev1"),
            ("1.0.0", "1.0.0-alpha1"),
            ("1.0.0", "1.0.0-alpha1"),
            ("1.0.0", "1.0.0-beta1"),
            ("1.0.0", "1.0.0-rc1"),
            ("1.0.0-alpha1", "1.0.0-dev1"),
            ("1.0.0-alpha1", "1.0.0-alpha1-dev1"),
            ("1.0.0-alpha2", "1.0.0-alpha1"),
            ("1.0.0-beta1", "1.0.0-dev1"),
            ("1.0.0-beta1", "1.0.0-alpha1"),
            ("1.0.0-beta1", "1.0.0-beta1-dev1"),
            ("1.0.0-beta2", "1.0.0-beta1"),
            ("1.0.0-rc1", "1.0.0-dev1"),
            ("1.0.0-rc1", "1.0.0-alpha1"),
            ("1.0.0-rc1", "1.0.0-beta1"),
            ("1.0.0-rc1", "1.0.0-rc1-dev1"),
            ("1.0.0-rc2", "1.0.0-rc1"),
        ]
        for version1, version2 in versions:
            self.assertTrue(
                Version.from_string(version1) > Version.from_string(version2),
                f"{version1} should be greater then {version2}",
            )

        versions = [
            ("1.0.0", "1.0.0"),
            ("1.0.0", "1.0.0+dev1"),
            ("1.0.0-dev1", "1.0.0-dev1"),
            ("1.0.0-dev1", "1.0.0-dev2"),
            ("1.0.0", "1.0.1"),
            ("1.0.0-dev1", "1.0.0"),
            ("1.0.0-dev1", "1.0.0-alpha1"),
            ("1.0.0-dev1", "1.0.0-beta1"),
            ("1.0.0-dev1", "1.0.0-rc1"),
            ("1.0.0+dev1", "1.0.0+dev1"),
            ("1.0.0+dev1", "1.0.0+dev2"),
            ("1.0.0-alpha1", "1.0.0-alpha1"),
            ("1.0.0-alpha1", "1.0.0-beta1"),
            ("1.0.0-alpha1", "1.0.0-rc1"),
            ("1.0.0-alpha1", "1.0.0-alpha1+dev1"),
            ("1.0.0-alpha1-dev1", "1.0.0-alpha1-dev1"),
            ("1.0.0-alpha1-dev1", "1.0.0-alpha1-dev2"),
            ("1.0.0-alpha1+dev1", "1.0.0-alpha1+dev2"),
            ("1.0.0-beta1", "1.0.0-rc1"),
            ("1.0.0-beta1", "1.0.0-beta1"),
            ("1.0.0-beta1", "1.0.0-beta1+dev1"),
            ("1.0.0-beta1-dev1", "1.0.0-beta1-dev1"),
            ("1.0.0-beta1-dev1", "1.0.0-beta1-dev2"),
            ("1.0.0-beta1+dev1", "1.0.0-beta1+dev2"),
            ("1.0.0-rc1", "1.0.0"),
            ("1.0.0-rc1", "1.0.0-rc1"),
            ("1.0.0-rc1", "1.0.0-rc1+dev1"),
            ("1.0.0-rc1-dev1", "1.0.0-rc1-dev1"),
            ("1.0.0-rc1-dev1", "1.0.0-rc1-dev2"),
            ("1.0.0-rc1+dev1", "1.0.0-rc1+dev2"),
        ]
        for version1, version2 in versions:
            self.assertFalse(
                Version.from_string(version1) > Version.from_string(version2),
                f"{version1} should not be greater then {version2}",
            )

    def test_is_dev_release(self):
        versions = [
            "1.0.0-dev1",
            "1.0.0-alpha1-dev1",
            "1.0.0-beta1-dev1",
            "1.0.0-rc1-dev1",
        ]
        for version in versions:
            self.assertTrue(
                Version.from_string(version).is_dev_release,
                f"{version} is not a dev release",
            )

        versions = [
            "1.0.0",
            "1.0.0+foo1",
            "1.0.0+dev1",
            "1.0.0-alpha1",
            "1.0.0-beta1",
            "1.0.0-rc1",
            "1.0.0-alpha1+dev1",
            "1.0.0-beta1+dev1",
            "1.0.0-rc1+dev1",
        ]
        for version in versions:
            self.assertFalse(
                Version.from_string(version).is_dev_release,
                f"{version} is a dev release",
            )

    def test_is_alpha_release(self):
        versions = [
            "1.0.0-alpha1",
            "1.0.0-alpha1+foo1",
            "1.0.0-alpha1-foo1",
            "1.0.0-alpha1+dev1",
            "1.0.0-alpha1-dev1",
        ]
        for version in versions:
            self.assertTrue(
                Version.from_string(version).is_alpha_release,
                f"{version} is not an alpha release",
            )

        versions = [
            "1.0.0",
            "1.0.0+dev1",
            "1.0.0-dev1",
            "1.0.0-a1",
            "1.0.0-b1",
            "1.0.0-rc1",
        ]
        for version in versions:
            self.assertFalse(
                Version.from_string(version).is_alpha_release,
                f"{version} is an alpha release",
            )

    def test_is_beta_release(self):
        versions = [
            "1.0.0-beta1",
            "1.0.0-beta1+foo1",
            "1.0.0-beta1-foo1",
            "1.0.0-beta1+dev1",
            "1.0.0-beta1-dev1",
        ]
        for version in versions:
            self.assertTrue(
                Version.from_string(version).is_beta_release,
                f"{version} is not a beta release",
            )

        versions = [
            "1.0.0",
            "1.0.0-dev1",
            "1.0.0+dev1",
            "1.0.0-alpha1",
            "1.0.0-b1",
            "1.0.0-rc1",
        ]
        for version in versions:
            self.assertFalse(
                Version.from_string(version).is_beta_release,
                f"{version} is a beta release",
            )

    def test_is_release_candidate(self):
        versions = [
            "1.0.0-rc1",
            "1.0.0-rc1+foo1",
            "1.0.0-rc1-foo1",
            "1.0.0-rc1+dev1",
            "1.0.0-rc1-dev1",
        ]
        for version in versions:
            self.assertTrue(
                Version.from_string(version).is_release_candidate,
                f"{version} is not a release candidate",
            )

        versions = [
            "1.0.0",
            "1.0.0+dev1",
            "1.0.0-dev1",
            "1.0.0-alpha1",
            "1.0.0-beta1",
        ]
        for version in versions:
            self.assertFalse(
                Version.from_string(version).is_release_candidate,
                f"{version} is a release candidate",
            )

    def test_pre(self):
        versions = [
            ("1.0.0", None),
            ("1.0.0+dev1", None),
            ("1.0.0-dev1", None),
            ("1.0.0-alpha1", ("alpha", 1)),
            ("1.0.0-beta1", ("beta", 1)),
            ("1.0.0-rc1", ("rc", 1)),
            ("1.0.0-rc1+foo1", ("rc", 1)),
            ("1.0.0-alpha1+dev1", ("alpha", 1)),
            ("1.0.0-beta1+dev1", ("beta", 1)),
            ("1.0.0-rc1+dev1", ("rc", 1)),
            ("1.0.0-alpha1-dev1", ("alpha", 1)),
            ("1.0.0-beta1-dev1", ("beta", 1)),
            ("1.0.0-rc1-dev1", ("rc", 1)),
        ]

        for version, expected in versions:
            self.assertEqual(Version.from_string(version).pre, expected)

    def test_local(self):
        versions = [
            ("1.0.0", None),
            ("1.0.0+foo1", ("foo", 1)),
            ("1.0.0+dev1", ("dev", 1)),
            ("1.0.0-dev1", None),
            ("1.0.0-alpha1", None),
            ("1.0.0-beta1", None),
            ("1.0.0-rc1", None),
            ("1.0.0-rc1+foo1", ("foo", 1)),
            ("1.0.0-alpha1+dev1", ("dev", 1)),
            ("1.0.0-beta1+dev1", ("dev", 1)),
            ("1.0.0-rc1+dev1", ("dev", 1)),
            ("1.0.0-alpha1-dev1", None),
            ("1.0.0-beta1-dev1", None),
            ("1.0.0-rc1-dev1", None),
        ]

        for version, expected in versions:
            local = Version.from_string(version).local
            self.assertEqual(
                local,
                expected,
                f"{version} has not the expected local {expected}. Instead it "
                f"is {local}",
            )


class SemanticVersionCalculatorTestCase(unittest.TestCase):
    def test_next_patch_version(self):
        calculator = VersionCalculator()

        versions = [
            ("0.0.1", "0.0.2"),
            ("1.2.3", "1.2.4"),
            ("1.2.3+dev1", "1.2.4"),
            ("1.2.3-dev1", "1.2.3"),
            ("1.2.3-foo1", "1.2.3"),
            ("1.2.3-alpha1", "1.2.3"),
            ("1.2.3-beta1", "1.2.3"),
            ("1.2.3-rc1", "1.2.3"),
            ("1.2.3-alpha1+dev1", "1.2.3"),
            ("1.2.3-beta1+dev1", "1.2.3"),
            ("1.2.3-rc1+dev1", "1.2.3"),
            ("1.2.3-alpha1-dev1", "1.2.3"),
            ("1.2.3-beta1-dev1", "1.2.3"),
            ("1.2.3-rc1-dev1", "1.2.3"),
            ("22.4.1", "22.4.2"),
            ("22.4.1+dev3", "22.4.2"),
            ("22.4.1-dev1", "22.4.1"),
            ("22.4.1-dev3", "22.4.1"),
            ("1.0.0-a1", "1.0.0"),
            ("1.1.0-alpha1", "1.1.0"),
            ("1.0.0+dev1", "1.0.1"),
            ("1.1.0+dev1", "1.1.1"),
            ("1.0.0-dev1", "1.0.0"),
            ("1.1.0-dev1", "1.1.0"),
        ]

        for current_version, assert_version in versions:
            release_version = calculator.next_patch_version(
                Version.from_string(current_version)
            )

            self.assertEqual(
                release_version,
                Version.from_string(assert_version),
                f"{release_version} is not the expected next patch version "
                f"{assert_version} for {current_version}",
            )

    def test_next_calendar_versions(self):
        calculator = VersionCalculator()
        today = datetime.today()
        year_short = today.year % 100

        current_versions = [
            "21.4.1-dev3",
            f"19.{today.month}.1-dev3",
            f"{year_short}.{today.month}.1-dev3",
            f"{year_short}.{today.month}.1",
        ]
        assert_versions = [
            f"{year_short}.{today.month}.0",
            f"{year_short}.{today.month}.0",
            f"{year_short}.{today.month}.1",
            f"{year_short}.{today.month}.2",
        ]

        for current_version, assert_version in zip(
            current_versions, assert_versions
        ):
            release_version = calculator.next_calendar_version(
                Version.from_string(current_version)
            )
            self.assertEqual(
                release_version, Version.from_string(assert_version)
            )

    def test_next_calendar_version_error(self):
        calculator = VersionCalculator()
        today = datetime.today()
        year_short = today.year % 100

        with self.assertRaisesRegex(VersionError, "'.+' is higher than '.+'."):
            calculator.next_calendar_version(
                Version.from_string(f"{year_short  + 1}.1.0")
            )

        with self.assertRaisesRegex(VersionError, "'.+' is higher than '.+'."):
            calculator.next_calendar_version(
                Version.from_string(f"{year_short}.{today.month + 1}.0")
            )

    def test_next_minor_version(self):
        calculator = VersionCalculator()

        versions = [
            ("0.0.1", "0.1.0"),
            ("1.2.3", "1.3.0"),
            ("1.2.3+dev1", "1.3.0"),
            ("1.2.3-dev1", "1.3.0"),
            ("1.2.3-foo1", "1.3.0"),
            ("1.2.3-alpha1", "1.3.0"),
            ("1.2.3-beta1", "1.3.0"),
            ("1.2.3-rc1", "1.3.0"),
            ("1.2.3-alpha1+dev1", "1.3.0"),
            ("1.2.3-beta1+dev1", "1.3.0"),
            ("1.2.3-rc1+dev1", "1.3.0"),
            ("22.4.1", "22.5.0"),
            ("22.4.1+dev3", "22.5.0"),
            ("22.4.1-dev1", "22.5.0"),
            ("22.4.1-dev3", "22.5.0"),
            ("1.0.0-a1", "1.0.0"),
            ("1.1.0-alpha1", "1.1.0"),
            ("1.0.0+dev1", "1.1.0"),
            ("1.1.0+dev1", "1.2.0"),
            ("1.0.0-dev1", "1.0.0"),
            ("1.1.0-dev1", "1.1.0"),
        ]
        for current_version, assert_version in versions:
            release_version = calculator.next_minor_version(
                Version.from_string(current_version)
            )
            self.assertEqual(
                release_version,
                Version.from_string(assert_version),
                f"{release_version} is not the expected next minor version "
                f"{assert_version} for {current_version}",
            )

    def test_next_major_version(self):
        calculator = VersionCalculator()

        versions = [
            ("0.0.1", "1.0.0"),
            ("1.2.3", "2.0.0"),
            ("1.2.3+dev1", "2.0.0"),
            ("1.2.3-dev1", "2.0.0"),
            ("1.2.3-foo1", "2.0.0"),
            ("1.2.3-alpha1", "2.0.0"),
            ("1.2.3-beta1", "2.0.0"),
            ("1.2.3-rc1", "2.0.0"),
            ("1.2.3-alpha1+dev1", "2.0.0"),
            ("1.2.3-beta1+dev1", "2.0.0"),
            ("1.2.3-rc1+dev1", "2.0.0"),
            ("1.2.3-alpha1-dev1", "2.0.0"),
            ("1.2.3-beta1-dev1", "2.0.0"),
            ("1.2.3-rc1-dev1", "2.0.0"),
            ("22.4.1", "23.0.0"),
            ("22.4.1+dev3", "23.0.0"),
            ("22.4.1-dev1", "23.0.0"),
            ("22.4.1-dev3", "23.0.0"),
            ("1.0.0-a1", "1.0.0"),
            ("1.0.0-beta1", "1.0.0"),
            ("1.1.0-alpha1", "2.0.0"),
            ("1.0.0+dev1", "2.0.0"),
            ("1.1.0+dev1", "2.0.0"),
            ("1.0.0-dev1", "1.0.0"),
            ("1.1.0-dev1", "2.0.0"),
        ]

        for current_version, assert_version in versions:
            release_version = calculator.next_major_version(
                Version.from_string(current_version)
            )
            self.assertEqual(
                release_version,
                Version.from_string(assert_version),
                f"{release_version} is not the expected next major version "
                f"{assert_version} for {current_version}",
            )

    def test_next_alpha_version(self):
        calculator = VersionCalculator()

        versions = [
            ("0.0.1", "0.0.2-alpha1"),
            ("1.2.3", "1.2.4-alpha1"),
            ("1.2.3+dev1", "1.2.4-alpha1"),
            ("1.2.3-dev1", "1.2.3-alpha1"),
            ("1.2.3-post1", "1.2.4-alpha1"),
            ("1.2.3-alpha1", "1.2.3-alpha2"),
            ("1.2.3-beta1", "1.2.4-alpha1"),
            ("1.2.3-rc1", "1.2.4-alpha1"),
            ("1.2.3-alpha1+dev1", "1.2.3-alpha2"),
            ("1.2.3-beta1+dev1", "1.2.4-alpha1"),
            ("1.2.3-rc1+dev1", "1.2.4-alpha1"),
            ("1.2.3-alpha1-dev1", "1.2.3-alpha1"),
            ("1.2.3-beta1-dev1", "1.2.4-alpha1"),
            ("1.2.3-beta1-dev1", "1.2.4-alpha1"),
            ("22.4.1", "22.4.2-alpha1"),
            ("22.4.1+dev3", "22.4.2-alpha1"),
            ("22.4.1-dev1", "22.4.1-alpha1"),
            ("22.4.1-dev3", "22.4.1-alpha1"),
            ("1.0.0-a1", "1.0.1-alpha1"),
            ("1.0.0-beta1", "1.0.1-alpha1"),
            ("1.1.0-alpha1", "1.1.0-alpha2"),
            ("1.0.0+dev1", "1.0.1-alpha1"),
            ("1.1.0+dev1", "1.1.1-alpha1"),
            ("1.0.0-dev1", "1.0.0-alpha1"),
            ("1.1.0-dev1", "1.1.0-alpha1"),
        ]

        for current_version, assert_version in versions:
            release_version = calculator.next_alpha_version(
                Version.from_string(current_version)
            )
            self.assertEqual(
                release_version,
                Version.from_string(assert_version),
                f"{release_version} is not the expected next alpha version "
                f"{assert_version} for {current_version}",
            )

    def test_next_beta_version(self):
        calculator = VersionCalculator()

        versions = [
            ("0.0.1", "0.0.2-beta1"),
            ("1.2.3", "1.2.4-beta1"),
            ("1.2.3+dev1", "1.2.4-beta1"),
            ("1.2.3-dev1", "1.2.3-beta1"),
            ("1.2.3-foo1", "1.2.4-beta1"),
            ("1.2.3-alpha1", "1.2.3-beta1"),
            ("1.2.3-beta1", "1.2.3-beta2"),
            ("1.2.3-rc1", "1.2.4-beta1"),
            ("1.2.3-alpha1+dev1", "1.2.3-beta1"),
            ("1.2.3-beta1+dev1", "1.2.3-beta2"),
            ("1.2.3-rc1+dev1", "1.2.4-beta1"),
            ("1.2.3-alpha1-dev1", "1.2.3-beta1"),
            ("1.2.3-beta1-dev1", "1.2.3-beta1"),
            ("1.2.3-rc1-dev1", "1.2.4-beta1"),
            ("22.4.1", "22.4.2-beta1"),
            ("22.4.1+dev3", "22.4.2-beta1"),
            ("22.4.1-dev1", "22.4.1-beta1"),
            ("22.4.1-dev3", "22.4.1-beta1"),
            # actually 1.0.0-beta1 would also be ok, but it would require to
            # to add extra code for not used versioning
            ("1.0.0-a1", "1.0.1-beta1"),
            ("1.0.0-beta1", "1.0.0-beta2"),
            ("1.1.0-alpha1", "1.1.0-beta1"),
            ("1.0.0+dev1", "1.0.1-beta1"),
            ("1.1.0+dev1", "1.1.1-beta1"),
            ("1.0.0-dev1", "1.0.0-beta1"),
            ("1.1.0-dev1", "1.1.0-beta1"),
        ]

        for current_version, assert_version in versions:
            release_version = calculator.next_beta_version(
                Version.from_string(current_version)
            )
            self.assertEqual(
                release_version,
                Version.from_string(assert_version),
                f"{release_version} is not the expected next beta version "
                f"{assert_version} for {current_version}",
            )

    def test_next_release_candidate_version(self):
        calculator = VersionCalculator()

        versions = [
            ("0.0.1", "0.0.2-rc1"),
            ("1.2.3", "1.2.4-rc1"),
            ("1.2.3+dev1", "1.2.4-rc1"),
            ("1.2.3-dev1", "1.2.3-rc1"),
            ("1.2.3-foo1", "1.2.4-rc1"),
            ("1.2.3-alpha1", "1.2.3-rc1"),
            ("1.2.3-beta1", "1.2.3-rc1"),
            ("1.2.3-rc1", "1.2.3-rc2"),
            ("1.2.3-alpha1+dev1", "1.2.3-rc1"),
            ("1.2.3-beta1+dev1", "1.2.3-rc1"),
            ("1.2.3-rc1+dev1", "1.2.3-rc2"),
            ("1.2.3-alpha1-dev1", "1.2.3-rc1"),
            ("1.2.3-beta1-dev1", "1.2.3-rc1"),
            ("1.2.3-rc1-dev1", "1.2.3-rc1"),
            ("22.4.1", "22.4.2-rc1"),
            ("22.4.1+dev3", "22.4.2-rc1"),
            ("22.4.1-dev1", "22.4.1-rc1"),
            ("22.4.1-dev3", "22.4.1-rc1"),
            ("1.0.0-a1", "1.0.1-rc1"),
            ("1.0.0-beta1", "1.0.0-rc1"),
            ("1.1.0-alpha1", "1.1.0-rc1"),
            ("1.0.0+dev1", "1.0.1-rc1"),
            ("1.1.0+dev1", "1.1.1-rc1"),
            ("1.0.0-dev1", "1.0.0-rc1"),
            ("1.1.0-dev1", "1.1.0-rc1"),
        ]

        for current_version, assert_version in versions:
            release_version = calculator.next_release_candidate_version(
                Version.from_string(current_version)
            )
            self.assertEqual(
                release_version,
                Version.from_string(assert_version),
                f"{release_version} is not the expected next rc version "
                f"{assert_version} for {current_version}",
            )

    def test_next_dev_version(self):
        calculator = VersionCalculator()

        versions = [
            ("0.0.1", "0.0.2-dev1"),
            ("1.2.3", "1.2.4-dev1"),
            ("1.2.3+dev1", "1.2.4-dev1"),
            ("1.2.3-dev1", "1.2.3-dev2"),
            ("1.2.3-foo1", "1.2.3-foo2-dev1"),
            ("1.2.3-alpha1", "1.2.3-alpha2-dev1"),
            ("1.2.3-beta1", "1.2.3-beta2-dev1"),
            ("1.2.3-rc1", "1.2.3-rc2-dev1"),
            ("1.2.3-alpha1+dev1", "1.2.3-alpha2-dev1"),
            ("1.2.3-beta1+dev1", "1.2.3-beta2-dev1"),
            ("1.2.3-rc1+dev1", "1.2.3-rc2-dev1"),
            ("1.2.3-alpha1-dev1", "1.2.3-alpha1-dev2"),
            ("1.2.3-beta1-dev1", "1.2.3-beta1-dev2"),
            ("1.2.3-rc1-dev1", "1.2.3-rc1-dev2"),
            ("22.4.1", "22.4.2-dev1"),
            ("22.4.1+dev3", "22.4.2-dev1"),
            ("22.4.1-dev1", "22.4.1-dev2"),
            ("22.4.1-dev3", "22.4.1-dev4"),
            ("1.0.0-a1", "1.0.0-a2-dev1"),
            ("1.0.0-beta1", "1.0.0-beta2-dev1"),
            ("1.1.0-alpha1", "1.1.0-alpha2-dev1"),
            ("1.0.0+dev1", "1.0.1-dev1"),
            ("1.1.0+dev1", "1.1.1-dev1"),
            ("1.0.0-dev1", "1.0.0-dev2"),
            ("1.1.0-dev1", "1.1.0-dev2"),
        ]

        for current_version, assert_version in versions:
            release_version = calculator.next_dev_version(
                Version.from_string(current_version)
            )
            self.assertEqual(
                release_version,
                Version.from_string(assert_version),
                f"{release_version} is not the expected next development "
                f"version {assert_version} for {current_version}",
            )



# Copyright (C) 2020-2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re
from pathlib import Path
from typing import Literal, Optional, Union

from lxml import etree

from ..errors import VersionError
from ..schemes import PEP440VersioningScheme
from ..version import Version, VersionUpdate
from ._command import VersionCommand

TEMPLATE = """# pylint: disable=invalid-name

# THIS IS AN AUTOGENERATED FILE. DO NOT TOUCH!

__version__ = "{}"\n"""

VERSION_PATTERN = (
    r"([0-9]+\.[0-9]+\.[0-9]+(-?([ab]|rc|alpha|beta)[0-9]+(.dev[0-9]+)?)?)"
)


def find_file(
    filename: Path, search_path: str, search_glob: str
) -> Optional[Path]:
    """Find a file somewhere within an directory tree

    Arg:
        filename: The file to look up
        search_path: The path to look for the file
        search_glob: The glob search pattern

    Returns:
        The file as Path object, if existing
    """
    search_path = Path(search_path).resolve()
    for file_path in search_path.glob(search_glob):
        if file_path.is_file() and file_path.name == filename.name:
            return file_path
    logging.warning("File %s not found in %s.", filename.name, search_path)
    return None


def replace_string_in_file(
    file_path: Path, pattern: str, replacement: str
) -> None:
    # Read the content of the file
    content = file_path.read_text(encoding="utf-8")

    # Search for the pattern in the content
    match = re.search(pattern, content)

    # Replace the matched group (Group 1) with the replacement
    if match:
        # Write the updated content back to the file
        file_path.write_text(
            content.replace(match.group(1), replacement), encoding="utf-8"
        )
    else:
        logging.warning(
            "Couldn't match the pattern %s in the content of %s.",
            pattern,
            file_path,
        )
        logging.warning("Content: %s", content)


# This class is used for Java Version command(s)
class JavaVersionCommand(VersionCommand):
    project_file_name = "pom.xml"
    _properties_file_path = Path("src/main/resources/application.properties")
    _pom_xml: Optional[etree.Element] = None

    def _get_version_from_pom_xml(self) -> Version:
        """
        Return the version information from the <version> tag of the
        pom.xml file. The version may be in non standardized form.
        """

        pom_xml: etree.Element = self.pom_xml

        version_element = pom_xml.find("{*}version")
        if version_element is None:
            raise VersionError("Version tag missing in pom.xml")

        return PEP440VersioningScheme.parse_version(version_element.text)

    def _update_pom_version(
        self,
        new_version: Version,
    ) -> None:
        """
        Update the version in the pom.xml file
        """
        pom_xml: etree.Element = self.pom_xml

        version_element = pom_xml.find("{*}version")
        if version_element is None:
            raise VersionError("Version tag missing in pom.xml")
        version_element.text = str(new_version)

        etree.ElementTree(pom_xml).write(
            self.project_file_path, pretty_print=True, encoding="utf-8"
        )

    def _update_properties_file(
        self,
        new_version: Version,
    ) -> None:
        # update the java properties file version
        if not self._properties_file_path.exists():
            # skip if not existing
            return
        pattern = rf"sentry\.release={VERSION_PATTERN}"
        replace_string_in_file(
            self._properties_file_path,
            pattern=pattern,
            replacement=str(new_version),
        )

    def _update_swagger_config(
        self,
        new_version: Version,
    ) -> None:
        # update swagger config file version
        swagger_config_file = find_file(
            filename=Path("SwaggerConfig.java"),
            search_path="src",
            search_glob="**/config/swagger/*",
        )
        if not swagger_config_file:
            # skip if not existing
            return
        pattern = rf'\.version\("{VERSION_PATTERN}"\)'
        replace_string_in_file(
            swagger_config_file, pattern=pattern, replacement=str(new_version)
        )

    @property
    def pom_xml(self) -> etree.Element:
        if self._pom_xml is not None:
            return self._pom_xml

        if not self.project_file_path.exists():
            raise VersionError("pom.xml file not found.")

        try:
            pom_xml: etree.ElementTree = etree.parse(self.project_file_path)
        except etree.XMLSyntaxError as e:
            raise VersionError(e) from e

        self._pom_xml = pom_xml.getroot()

        return self._pom_xml

    def get_current_version(self) -> Version:
        """Get the current version of this project
        In go the version is only defined within the repository
        tags, thus we need to check git, what tag is the latest"""
        return self._get_version_from_pom_xml()

    def verify_version(
        self, version: Union[Literal["current"], Version, None]
    ) -> None:
        """Verify the current version of this project"""
        current_version = self.get_current_version()

        if current_version != version:
            raise VersionError(
                f"Provided version {version} does not match the "
                f"current version {current_version} in "
                f"{self.project_file_path}."
            )

    def update_version(
        self, new_version: Version, *, force: bool = False
    ) -> VersionUpdate:
        try:
            package_version = self.get_current_version()
            if not force and new_version == package_version:
                return VersionUpdate(previous=package_version, new=new_version)
        except VersionError:
            # just ignore current version and override it
            package_version = None

        changed_files = [self.project_file_path]
        self._update_pom_version(new_version=new_version)
        self._update_properties_file(new_version=new_version)
        self._update_swagger_config(new_version=new_version)

        return VersionUpdate(
            previous=package_version,
            new=new_version,
            changed_files=changed_files,
        )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .core import ActionIO, Console
from .env import GitHubEnvironment
from .errors import GitHubActionsError
from .event import (
    GitHubEvent,
    GitHubPullRequestEvent,
    Label,
    PullRequestState,
    Ref,
)
from .main import main

__all__ = (
    "GitHubActionsError",
    "ActionIO",
    "Console",
    "GitHubEnvironment",
    "GitHubEvent",
    "Label",
    "Ref",
    "PullRequestState",
    "GitHubPullRequestEvent",
    "main",
)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import AsyncContextManager, AsyncIterator, Optional, Union

from pontos.github.api.client import GitHubAsyncREST, Params
from pontos.github.models.artifact import Artifact
from pontos.helper import AsyncDownloadProgressIterable, download_async


class GitHubAsyncRESTArtifacts(GitHubAsyncREST):
    def _get_paged_artifacts(
        self, api, *, params: Optional[Params] = None
    ) -> AsyncIterator[Artifact]:
        return self._get_paged_items(
            api, "artifacts", Artifact, params=params  # type: ignore
        )

    def get_all(self, repo: str) -> AsyncIterator[Artifact]:
        """
        List all artifacts of a repository

        https://docs.github.com/en/rest/actions/artifacts#list-artifacts-for-a-repository

        Args:
            repo: GitHub repository (owner/name) to use

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            An async iterator for the received artifacts

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    async for artifact in api.artifacts.get_all("foo/bar"):
                        print(artifact)
        """
        api = f"/repos/{repo}/actions/artifacts"
        return self._get_paged_artifacts(api)

    async def get(self, repo: str, artifact: Union[str, int]) -> Artifact:
        """
        Get a single artifact of a repository

        https://docs.github.com/en/rest/actions/artifacts#get-an-artifact

        Args:
            repo: GitHub repository (owner/name) to use
            artifact: ID of the artifact

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the artifact

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    artifact = await api.artifacts.get("foo/bar", 123)
                    print(artifact)
        """
        api = f"/repos/{repo}/actions/artifacts/{artifact}"
        response = await self._client.get(api)
        response.raise_for_status()
        return Artifact.from_dict(response.json())

    def get_workflow_run_artifacts(
        self, repo: str, run: Union[str, int]
    ) -> AsyncIterator[Artifact]:
        # pylint: disable=line-too-long
        """
        List all artifacts for a workflow run

        https://docs.github.com/en/rest/actions/artifacts#list-workflow-run-artifacts

        Args:
            repo: GitHub repository (owner/name) to use
            run: The unique identifier of the workflow run

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            An async iterator for the received artifacts

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    async for artifact in api.artifacts.get_workflow_run_artifacts("foo/bar", 234):
                        print(artifact)
        """  # noqa: E501
        api = f"/repos/{repo}/actions/runs/{run}/artifacts"
        return self._get_paged_artifacts(api)

    async def delete(self, repo: str, artifact: Union[str, int]) -> None:
        """
        Delete an artifact of a repository

        https://docs.github.com/en/rest/actions/artifacts#delete-an-artifact

        Args:
            repo: GitHub repository (owner/name) to use
            artifact: ID of the artifact

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    await api.artifacts.delete("foo/bar", 123):
        """
        api = f"/repos/{repo}/actions/artifacts/{artifact}"
        response = await self._client.delete(api)
        response.raise_for_status()

    def download(
        self, repo: str, artifact: Union[str, int]
    ) -> AsyncContextManager[AsyncDownloadProgressIterable[bytes]]:
        """
        Download a repository artifact zip file

        https://docs.github.com/en/rest/actions/artifacts#download-an-artifact

        Args:
            repo: GitHub repository (owner/name) to use
            artifact: ID of the artifact

        Raises:
            HTTPStatusError: If the request was invalid

        Example:
            .. code-block:: python

            from pontos.github.api import GitHubAsyncRESTApi

            async with GitHubAsyncRESTApi("...") as api:
                print("Downloading", end="")

                with Path("foo.baz").open("wb") as f:
                    async with api.artifacts.download(
                        "org/repo", "123"
                    ) as download:
                        async for content, progress in download:
                            f.write(content)
                            print(".", end="")
        """
        api = f"/repos/{repo}/actions/artifacts/{artifact}/zip"
        return download_async(self._client.stream(api))



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=line-too-long

import unittest
from pathlib import Path
from unittest.mock import MagicMock

from httpx import HTTPStatusError

from pontos.github.api.branch import (
    GitHubAsyncRESTBranches,
    update_from_applied_settings,
)
from pontos.github.models.branch import BranchProtection
from tests.github.api import GitHubAsyncRESTTestCase, create_response

here = Path(__file__).parent


class UpdateFromAppliedSettingsTestCase(unittest.TestCase):
    def test_update_from_applied_settings(self):
        branch_protection = BranchProtection.from_dict(
            {
                "url": "https://api.github.com/repos/foo/bar/branches/branch_protection/protection/restrictions",
                "users_url": "https://api.github.com/repos/foo/bar/branches/branch_protection/protection/restrictions/users",
                "teams_url": "https://api.github.com/repos/foo/bar/branches/branch_protection/protection/restrictions/teams",
                "apps_url": "https://api.github.com/repos/foo/bar/branches/branch_protection/protection/restrictions/apps",
                "users": [
                    {
                        "login": "greenbonebot",
                        "id": 123,
                        "node_id": "MDQ6VXNlcjg1MjU0NjY2",
                        "avatar_url": "https://avatars.githubusercontent.com/u/85254666?v=4",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/greenbonebot",
                        "html_url": "https://github.com/greenbonebot",
                        "followers_url": "https://api.github.com/users/greenbonebot/followers",
                        "following_url": "https://api.github.com/users/greenbonebot/following{/other_user}",
                        "gists_url": "https://api.github.com/users/greenbonebot/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/greenbonebot/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/greenbonebot/subscriptions",
                        "organizations_url": "https://api.github.com/users/greenbonebot/orgs",
                        "repos_url": "https://api.github.com/users/greenbonebot/repos",
                        "events_url": "https://api.github.com/users/greenbonebot/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/greenbonebot/received_events",
                        "type": "User",
                        "site_admin": False,
                    }
                ],
                "teams": [],
                "apps": [],
                "lock_branch": {"enabled": False},
                "allow_fork_syncing": {"enabled": False},
                "required_signatures": {"enabled": False},
            }
        )

        kwargs = update_from_applied_settings(
            branch_protection=branch_protection,
            lock_branch=True,
            allow_fork_syncing=True,
            allow_deletions=True,
            allow_force_pushes=True,
        )

        self.assertTrue(kwargs["lock_branch"])
        self.assertTrue(kwargs["allow_fork_syncing"])
        self.assertTrue(kwargs["allow_deletions"])
        self.assertTrue(kwargs["allow_force_pushes"])

        self.assertFalse(kwargs["required_signatures"])

        self.assertIsNone(kwargs["required_linear_history"])
        self.assertIsNone(kwargs["block_creations"])
        self.assertIsNone(kwargs["required_conversation_resolution"])
        self.assertIsNone(kwargs["enforce_admins"])


class GitHubAsyncRESTBranchesTestCase(GitHubAsyncRESTTestCase):
    api_cls = GitHubAsyncRESTBranches

    async def test_exists(self):
        response = create_response(is_success=True)
        self.client.get.return_value = response

        self.assertTrue(await self.api.exists("foo/bar", "baz"))
        self.client.get.assert_awaited_once_with("/repos/foo/bar/branches/baz")

    async def test_not_exists(self):
        response = create_response(is_success=False)
        self.client.get.return_value = response

        self.assertFalse(await self.api.exists("foo/bar", "baz"))
        self.client.get.assert_awaited_once_with("/repos/foo/bar/branches/baz")

    async def test_delete_branch(self):
        response = create_response()
        self.client.delete.return_value = response

        await self.api.delete("foo/bar", "baz")

        self.client.delete.assert_awaited_once_with(
            "/repos/foo/bar/git/refs/baz"
        )

    async def test_delete_branch_failure(self):
        response = create_response()
        error = HTTPStatusError("404", request=MagicMock(), response=response)
        response.raise_for_status.side_effect = error

        self.client.delete.return_value = response

        with self.assertRaises(HTTPStatusError):
            await self.api.delete("foo/bar", "baz")

        self.client.delete.assert_awaited_once_with(
            "/repos/foo/bar/git/refs/baz"
        )

    async def test_protection_rules(self):
        rules = {
            "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
            "required_signatures": {
                "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection/required_signatures",
                "enabled": False,
            },
            "enforce_admins": {
                "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection/enforce_admins",
                "enabled": False,
            },
        }
        response = create_response()
        response.json.return_value = rules

        self.client.get.return_value = response

        data = await self.api.protection_rules("foo/bar", "baz")

        self.client.get.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection"
        )
        self.assertEqual(
            data.url,
            "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        )

    async def test_protection_rules_failure(self):
        response = create_response()
        error = HTTPStatusError("404", request=MagicMock(), response=response)
        response.raise_for_status.side_effect = error

        self.client.get.return_value = response

        with self.assertRaises(HTTPStatusError):
            await self.api.protection_rules("foo/bar", "baz")

        self.client.get.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection"
        )

    async def test_delete_protection_rules(self):
        response = create_response()
        self.client.delete.return_value = response

        await self.api.delete_protection_rules("foo/bar", "baz")

        self.client.delete.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection"
        )

    async def test_delete_protection_rules_failure(self):
        response = create_response()
        error = HTTPStatusError("404", request=MagicMock(), response=response)
        response.raise_for_status.side_effect = error

        self.client.delete.return_value = response

        with self.assertRaises(HTTPStatusError):
            await self.api.delete_protection_rules("foo/bar", "baz")

        self.client.delete.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection"
        )

    async def test_update_protection_rules_defaults(self):
        response = create_response()
        response.json.return_value = {
            "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        }
        self.client.put.return_value = response

        rules = await self.api.update_protection_rules(
            "foo/bar",
            "baz",
        )
        self.client.put.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection",
            data={
                "required_status_checks": None,
                "enforce_admins": None,
                "required_pull_request_reviews": None,
                "restrictions": None,
            },
        )

        self.assertEqual(
            rules.url,
            "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        )

    async def test_update_protection_rules(self):
        response = create_response()
        response.json.return_value = {
            "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        }
        self.client.put.return_value = response

        rules = await self.api.update_protection_rules(
            "foo/bar",
            "baz",
            required_status_checks=[("foo", "123"), ("bar", None)],
            require_branches_to_be_up_to_date=True,
            enforce_admins=True,
            dismissal_restrictions_users=("foo", "bar"),
            dismissal_restrictions_teams=("team_foo", "team_bar"),
            dismissal_restrictions_apps=("123", "321"),
            dismiss_stale_reviews=True,
            require_code_owner_reviews=True,
            required_approving_review_count=2,
            require_last_push_approval=True,
            bypass_pull_request_allowances_users=("foo", "bar"),
            bypass_pull_request_allowances_teams=("team_foo", "team_bar"),
            bypass_pull_request_allowances_apps=("123", "321"),
            restrictions_users=("foo", "bar"),
            restrictions_teams=("team_foo", "team_bar"),
            restrictions_apps=("123", "321"),
            required_linear_history=True,
            allow_force_pushes=True,
            allow_deletions=True,
            block_creations=True,
            required_conversation_resolution=True,
            lock_branch=True,
            allow_fork_syncing=True,
        )

        self.client.put.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection",
            data={
                "required_status_checks": {
                    "strict": True,
                    "checks": [
                        {"context": "foo", "app_id": "123"},
                        {"context": "bar"},
                    ],
                },
                "enforce_admins": True,
                "required_pull_request_reviews": {
                    "dismissal_restrictions": {
                        "users": ["foo", "bar"],
                        "teams": ["team_foo", "team_bar"],
                        "apps": ["123", "321"],
                    },
                    "dismiss_stale_reviews": True,
                    "require_code_owner_reviews": True,
                    "required_approving_review_count": 2,
                    "require_last_push_approval": True,
                    "bypass_pull_request_allowances": {
                        "users": ["foo", "bar"],
                        "teams": ["team_foo", "team_bar"],
                        "apps": ["123", "321"],
                    },
                },
                "restrictions": {
                    "users": ["foo", "bar"],
                    "teams": ["team_foo", "team_bar"],
                    "apps": ["123", "321"],
                },
                "required_linear_history": True,
                "allow_force_pushes": True,
                "allow_deletions": True,
                "block_creations": True,
                "required_conversation_resolution": True,
                "lock_branch": True,
                "allow_fork_syncing": True,
            },
        )

        self.assertEqual(
            rules.url,
            "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        )

    async def test_update_protection_defaults(self):
        response = create_response()
        response.json.return_value = {
            "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        }
        self.client.put.return_value = response

        rules = await self.api.update_protection_rules(
            "foo/bar",
            "baz",
        )

        self.client.put.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection",
            data={
                "required_status_checks": None,
                "enforce_admins": None,
                "required_pull_request_reviews": None,
                "restrictions": None,
            },
        )

        self.assertEqual(
            rules.url,
            "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        )

    async def test_update_protection_rules_up_to_date_branch(self):
        response = create_response()
        response.json.return_value = {
            "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        }
        self.client.put.return_value = response

        rules = await self.api.update_protection_rules(
            "foo/bar",
            "baz",
            require_branches_to_be_up_to_date=True,
        )

        self.client.put.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection",
            data={
                "required_status_checks": {
                    "strict": True,
                    "checks": [],
                },
                "enforce_admins": None,
                "required_pull_request_reviews": None,
                "restrictions": None,
            },
        )

        self.assertEqual(
            rules.url,
            "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        )

    async def test_update_protection_rules_restriction_users(self):
        response = create_response()
        response.json.return_value = {
            "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        }
        self.client.put.return_value = response

        rules = await self.api.update_protection_rules(
            "foo/bar", "baz", restrictions_users=["foo", "bar"]
        )

        self.client.put.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection",
            data={
                "enforce_admins": None,
                "required_pull_request_reviews": None,
                "required_status_checks": None,
                "restrictions": {"users": ["foo", "bar"], "teams": []},
            },
        )

        self.assertEqual(
            rules.url,
            "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        )

    async def test_update_protection_rules_restriction_teams(self):
        response = create_response()
        response.json.return_value = {
            "url": "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        }
        self.client.put.return_value = response

        rules = await self.api.update_protection_rules(
            "foo/bar", "baz", restrictions_teams=["foo", "bar"]
        )

        self.client.put.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection",
            data={
                "enforce_admins": None,
                "required_pull_request_reviews": None,
                "required_status_checks": None,
                "restrictions": {"teams": ["foo", "bar"], "users": []},
            },
        )

        self.assertEqual(
            rules.url,
            "https://api.github.com/repos/octocat/Hello-World/branches/main/protection",
        )

    async def test_update_protection_rules_failure(self):
        response = create_response()
        error = HTTPStatusError("404", request=MagicMock(), response=response)
        response.raise_for_status.side_effect = error

        self.client.put.return_value = response

        with self.assertRaises(HTTPStatusError):
            await self.api.update_protection_rules(
                "foo/bar",
                "baz",
                enforce_admins=True,
            )

        self.client.put.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection",
            data={
                "enforce_admins": True,
                "required_status_checks": None,
                "required_pull_request_reviews": None,
                "restrictions": None,
            },
        )

    async def test_enable_enforce_admins(self):
        response = create_response()
        self.client.post.return_value = response

        await self.api.set_enforce_admins("foo/bar", "baz", enforce_admins=True)

        self.client.post.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection/enforce_admins"
        )

    async def test_disable_enforce_admins(self):
        response = create_response()
        self.client.delete.return_value = response

        await self.api.set_enforce_admins(
            "foo/bar", "baz", enforce_admins=False
        )

        self.client.delete.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection/enforce_admins"
        )

    async def test_enable_required_signatures(self):
        response = create_response()
        self.client.post.return_value = response

        await self.api.set_required_signatures(
            "foo/bar", "baz", required_signatures=True
        )

        self.client.post.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection/required_signatures"
        )

    async def test_disable_required_signatures(self):
        response = create_response()
        self.client.delete.return_value = response

        await self.api.set_required_signatures(
            "foo/bar", "baz", required_signatures=False
        )

        self.client.delete.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection/required_signatures"
        )

    async def test_update_required_status_checks(self):
        response = create_response()
        self.client.patch.return_value = response

        await self.api.update_required_status_checks(
            "foo/bar",
            "baz",
            required_status_checks=[("foo", "123"), ("bar", None)],
            require_branches_to_be_up_to_date=True,
        )

        self.client.patch.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection/required_status_checks",
            data={
                "strict": True,
                "checks": [
                    {"context": "foo", "app_id": "123"},
                    {"context": "bar"},
                ],
            },
        )

    async def test_remove_required_required_status_checks(self):
        response = create_response()
        self.client.delete.return_value = response

        await self.api.remove_required_status_checks("foo/bar", "baz")

        self.client.delete.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection/required_status_checks"
        )

    async def test_remove_required_required_status_checks_failure(self):
        response = create_response()
        error = HTTPStatusError("404", request=MagicMock(), response=response)
        response.raise_for_status.side_effect = error

        self.client.delete.return_value = response

        with self.assertRaises(HTTPStatusError):
            await self.api.remove_required_status_checks("foo/bar", "baz")

        self.client.delete.assert_awaited_once_with(
            "/repos/foo/bar/branches/baz/protection/required_status_checks"
        )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-lines, redefined-builtin, line-too-long
# ruff: noqa: E501

from datetime import datetime, timezone
from unittest.mock import MagicMock

import httpx

from pontos.github.api.tags import GitHubAsyncRESTTags
from pontos.github.models.tag import GitObjectType, Tag, VerificationReason
from tests import AsyncIteratorMock, aiter, anext
from tests.github.api import GitHubAsyncRESTTestCase, create_response

TAG_JSON = {
    "node_id": "MDM6VGFnOTQwYmQzMzYyNDhlZmFlMGY5ZWU1YmM3YjJkNWM5ODU4ODdiMTZhYw==",
    "tag": "v0.0.1",
    "sha": "940bd336248efae0f9ee5bc7b2d5c985887b16ac",
    "url": "https://api.github.com/repos/octocat/Hello-World/git/tags/940bd336248efae0f9ee5bc7b2d5c985887b16ac",
    "message": "initial version",
    "tagger": {
        "name": "Monalisa Octocat",
        "email": "octocat@github.com",
        "date": "2014-11-07T22:01:45Z",
    },
    "object": {
        "type": "commit",
        "sha": "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
        "url": "https://api.github.com/repos/octocat/Hello-World/git/commits/c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
    },
    "verification": {
        "verified": False,
        "reason": "unsigned",
        "signature": None,
        "payload": None,
    },
}


TAGS_JSON = [
    {
        "name": "v0.1.0",
        "zipball_url": "https://api.github.com/repos/foo/ui/zipball/refs/tags/v0.1.0",
        "tarball_url": "https://api.github.com/repos/foo/ui/tarball/refs/tags/v0.1.0",
        "commit": {
            "sha": "b55408280162c20cf0ffef155618270d26fe7352",
            "url": "https://api.github.com/repos/foo/ui/commits/b55408280162c20cf0ffef155618270d26fe7352",
        },
        "node_id": "REF_kwDOHeuT1rByZWZzL3RhZ3MvdjAuMS4w",
    },
    {
        "name": "0.0.72",
        "zipball_url": "https://api.github.com/repos/foo/ui/zipball/refs/tags/0.0.72",
        "tarball_url": "https://api.github.com/repos/foo/ui/tarball/refs/tags/0.0.72",
        "commit": {
            "sha": "fd54097278df75332917d9d1a2d940083513a6b2",
            "url": "https://api.github.com/repos/foo/ui/commits/fd54097278df75332917d9d1a2d940083513a6b2",
        },
        "node_id": "REF_kwDOHeuT1rByZWZzL3RhZ3MvMC4wLjcy",
    },
]


class GitHubAsyncRESTTagsTestCase(GitHubAsyncRESTTestCase):
    api_cls = GitHubAsyncRESTTags

    def assertTag(self, tag: Tag):  # pylint: disable=invalid-name
        self.assertEqual(
            tag.node_id,
            "MDM6VGFnOTQwYmQzMzYyNDhlZmFlMGY5ZWU1YmM3YjJkNWM5ODU4ODdiMTZhYw==",
        )
        self.assertEqual(tag.tag, "v0.0.1")
        self.assertEqual(tag.sha, "940bd336248efae0f9ee5bc7b2d5c985887b16ac")
        self.assertEqual(
            tag.url,
            "https://api.github.com/repos/octocat/Hello-World/git/tags/940bd336248efae0f9ee5bc7b2d5c985887b16ac",
        )
        self.assertEqual(tag.message, "initial version")

        tagger = tag.tagger
        self.assertEqual(tagger.name, "Monalisa Octocat")
        self.assertEqual(tagger.email, "octocat@github.com")
        self.assertEqual(
            tagger.date, datetime(2014, 11, 7, 22, 1, 45, tzinfo=timezone.utc)
        )

        tag_object = tag.object
        self.assertEqual(tag_object.type, GitObjectType.COMMIT)
        self.assertEqual(
            tag_object.sha, "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c"
        )
        self.assertEqual(
            tag_object.url,
            "https://api.github.com/repos/octocat/Hello-World/git/commits/c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
        )

        verification = tag.verification
        self.assertFalse(verification.verified)
        self.assertEqual(verification.reason, VerificationReason.UNSIGNED)
        self.assertIsNone(verification.payload)
        self.assertIsNone(verification.signature)

    async def test_create(self):
        response = create_response()
        response.json.return_value = TAG_JSON
        self.client.post.return_value = response

        tag = await self.api.create(
            "octocat/Hello-World",
            "v0.0.1",
            "initial version",
            "Monalisa Octocat",
            "octocat@github.com",
            "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
        )

        self.client.post.assert_awaited_once_with(
            "/repos/octocat/Hello-World/git/tags",
            data={
                "tag": "v0.0.1",
                "message": "initial version",
                "object": "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
                "type": "commit",
                "tagger": {
                    "name": "Monalisa Octocat",
                    "email": "octocat@github.com",
                },
            },
        )

        self.assertTag(tag)

    async def test_create_failure(self):
        response = create_response()
        self.client.post.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.create(
                "octocat/Hello-World",
                "v0.0.1",
                "initial version",
                "Monalisa Octocat",
                "octocat@github.com",
                "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
            )

        self.client.post.assert_awaited_once_with(
            "/repos/octocat/Hello-World/git/tags",
            data={
                "tag": "v0.0.1",
                "message": "initial version",
                "object": "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
                "type": "commit",
                "tagger": {
                    "name": "Monalisa Octocat",
                    "email": "octocat@github.com",
                },
            },
        )

    async def test_create_tag_reference(self):
        response = create_response()
        self.client.post.return_value = response

        await self.api.create_tag_reference(
            "foo/bar", "v1.0.0", "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c"
        )

        self.client.post.assert_awaited_once_with(
            "/repos/foo/bar/git/refs",
            data={
                "ref": "refs/tags/v1.0.0",
                "sha": "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
            },
        )

    async def test_create_tag_reference_failure(self):
        response = create_response()
        self.client.post.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.create_tag_reference(
                "foo/bar", "v1.0.0", "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c"
            )

        self.client.post.assert_awaited_once_with(
            "/repos/foo/bar/git/refs",
            data={
                "ref": "refs/tags/v1.0.0",
                "sha": "c3d0be41ecbe669545ee3e94d31ed9a4bc91ee3c",
            },
        )

    async def test_get(self):
        response = create_response()
        response.json.return_value = TAG_JSON
        self.client.get.return_value = response

        tag = await self.api.get("octocat/Hello-World", "v0.0.1")

        self.client.get.assert_awaited_once_with(
            "/repos/octocat/Hello-World/git/tags/v0.0.1"
        )

        self.assertTag(tag)

    async def test_get_failure(self):
        response = create_response()
        self.client.get.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.get("octocat/Hello-World", "v0.0.1")

        self.client.get.assert_awaited_once_with(
            "/repos/octocat/Hello-World/git/tags/v0.0.1"
        )

    async def test_get_all(self):
        response = create_response()
        response.json.return_value = TAGS_JSON
        self.client.get_all.return_value = AsyncIteratorMock([response])

        async_it = aiter(self.api.get_all("foo"))

        tag = await anext(async_it)
        self.assertEqual(tag.name, "v0.1.0")
        self.assertEqual(
            tag.commit.sha, "b55408280162c20cf0ffef155618270d26fe7352"
        )
        self.assertEqual(
            tag.commit.url,
            "https://api.github.com/repos/foo/ui/commits/b55408280162c20cf0ffef155618270d26fe7352",
        )
        self.assertEqual(
            tag.zipball_url,
            "https://api.github.com/repos/foo/ui/zipball/refs/tags/v0.1.0",
        )
        self.assertEqual(
            tag.tarball_url,
            "https://api.github.com/repos/foo/ui/tarball/refs/tags/v0.1.0",
        )
        self.assertEqual(tag.node_id, "REF_kwDOHeuT1rByZWZzL3RhZ3MvdjAuMS4w")
        tag2 = await anext(async_it)
        self.assertEqual(tag2.name, "0.0.72")



# Copyright (C) 2020-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
from argparse import (
    ArgumentParser,
    ArgumentTypeError,
    BooleanOptionalAction,
    Namespace,
)
from enum import Enum
from pathlib import Path
from typing import Callable, Optional, Tuple, Type

from pontos.release.helper import ReleaseType
from pontos.release.show import OutputFormat, show
from pontos.version.schemes import (
    VERSIONING_SCHEMES,
    PEP440VersioningScheme,
    VersioningScheme,
    versioning_scheme_argument_type,
)

from .create import create_release
from .sign import sign

DEFAULT_SIGNING_KEY = "0ED1E580"


def to_choices(enum: Type[Enum]) -> str:
    return ", ".join([t.value for t in enum])


def enum_type(enum: Type[Enum]) -> Callable[[str], Enum]:
    def convert(value: str) -> Enum:
        try:
            return enum(value)
        except ValueError:
            raise ArgumentTypeError(
                f"invalid value {value}. Expected one of {to_choices(enum)}."
            ) from None

    return convert


class ReleaseVersionAction(
    argparse._StoreAction
):  # pylint: disable=protected-access
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, "release_type", ReleaseType.VERSION)
        setattr(namespace, self.dest, values)


def parse_args(args) -> Tuple[Optional[str], Optional[str], Namespace]:
    """
    Return user, token, parsed arguments
    """
    parser = ArgumentParser(
        description="Release handling utility.",
        prog="pontos-release",
    )

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Don't print messages to the terminal",
    )

    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        help="additional help",
        dest="command",
        required=True,
    )

    create_parser = subparsers.add_parser(
        "create", help="Create a new release", aliases=["release"]
    )
    create_parser.set_defaults(func=create_release)
    create_parser.add_argument(
        "--versioning-scheme",
        help="Versioning scheme to use for parsing and handling version "
        f"information. Choices are {', '.join(VERSIONING_SCHEMES.keys())}. "
        "Default: %(default)s",
        default="pep440",
        type=versioning_scheme_argument_type,
    )
    create_parser.add_argument(
        "--release-type",
        help="Select the release type for calculating the release version. "
        f"Possible choices are: {to_choices(ReleaseType)}.",
        type=enum_type(ReleaseType),
    )
    create_parser.add_argument(
        "--release-version",
        help=(
            "Will release changelog as version. "
            "Default: lookup version in project definition."
        ),
        action=ReleaseVersionAction,
    )
    create_parser.add_argument(
        "--release-series",
        help="Create a release for a release series. Setting a release series "
        "is required if the latest tag version is newer then the to be "
        'released version. Examples: "1.2", "2", "22.4"',
    )

    create_parser.add_argument(
        "--next-version",
        help=(
            "Sets the next version in project definition "
            "after the release. Default: set to next dev version"
        ),
    )

    create_parser.add_argument(
        "--git-remote-name",
        help="The git remote name to push the commits and tag to",
    )
    create_parser.add_argument(
        "--git-tag-prefix",
        default="v",
        const="",
        nargs="?",
        help="Prefix for git tag versions. Default: %(default)s",
    )
    create_parser.add_argument(
        "--git-signing-key",
        help="The key to sign the commits and tag for a release",
        default=os.environ.get("GPG_SIGNING_KEY"),
    )
    create_parser.add_argument(
        "--project",
        help="The github project",
    )
    create_parser.add_argument(
        "--space",
        default="greenbone",
        help="User/Team name in github",
    )
    create_parser.add_argument(
        "--local",
        action="store_true",
        help="Only create release changes locally and do not upload them to a "
        "remote repository. Also do not create a GitHub release.",
    )
    create_parser.add_argument(
        "--conventional-commits-config",
        dest="cc_config",
        type=Path,
        help="Conventional commits config file (toml), including conventions."
        " If not set defaults are used.",
    )
    create_parser.add_argument(
        "--update-project",
        help="Update version in project files like pyproject.toml. By default "
        "project files are updated.",
        action=BooleanOptionalAction,
        default=True,
    )

    sign_parser = subparsers.add_parser(
        "sign", help="Create signatures for an existing release"
    )
    sign_parser.set_defaults(func=sign)
    sign_parser.add_argument(
        "--signing-key",
        default=DEFAULT_SIGNING_KEY,
        help="The key to sign zip, tarballs of a release. Default %(default)s.",
    )
    sign_parser.add_argument(
        "--versioning-scheme",
        help="Versioning scheme to use for parsing and handling version "
        f"information. Choices are {', '.join(VERSIONING_SCHEMES.keys())}. "
        "Default: %(default)s",
        default="pep440",
        type=versioning_scheme_argument_type,
    )
    sign_parser.add_argument(
        "--release-version",
        help="Will release changelog as version. Must be PEP 440 compliant.",
    )
    sign_parser.add_argument(
        "--release-series",
        help="Sign release files for a release series. Setting a release "
        "series is required if the latest tag version is newer then the to be "
        'signed version. Examples: "1.2", "2", "22.4"',
    )
    sign_parser.add_argument(
        "--git-tag-prefix",
        default="v",
        const="",
        nargs="?",
        help="Prefix for git tag versions. Default: %(default)s",
    )
    sign_parser.add_argument(
        "--project",
        help="The github project",
    )
    sign_parser.add_argument(
        "--space",
        default="greenbone",
        help="user/team name in github",
    )

    sign_parser.add_argument(
        "--passphrase",
        help=(
            "Use gpg in a headless mode e.g. for "
            "the CI and use this passphrase for signing."
        ),
    )
    sign_parser.add_argument(
        "--dry-run", action="store_true", help="Do not upload signed files."
    )

    show_parser = subparsers.add_parser(
        "show",
        help="Show release information about the current release version and "
        "determine the next release version",
    )
    show_parser.set_defaults(func=show)
    show_parser.add_argument(
        "--versioning-scheme",
        help="Versioning scheme to use for parsing and handling version "
        f"information. Choices are {', '.join(VERSIONING_SCHEMES.keys())}. "
        "Default: %(default)s",
        default="pep440",
        type=versioning_scheme_argument_type,
    )
    show_parser.add_argument(
        "--release-type",
        help="Select the release type for calculating the release version. "
        f"Possible choices are: {to_choices(ReleaseType)}.",
        type=enum_type(ReleaseType),
    )
    show_parser.add_argument(
        "--release-version",
        help=(
            "Will release changelog as version. "
            "Default: lookup version in project definition."
        ),
        action=ReleaseVersionAction,
    )
    show_parser.add_argument(
        "--release-series",
        help="Create a release for a release series. Setting a release series "
        "is required if the latest tag version is newer then the to be "
        'released version. Examples: "1.2", "2", "22.4"',
    )
    show_parser.add_argument(
        "--git-tag-prefix",
        default="v",
        const="",
        nargs="?",
        help="Prefix for git tag versions. Default: %(default)s",
    )
    show_parser.add_argument(
        "--output-format",
        help="Print in the desired output format. "
        f"Possible choices are: {to_choices(OutputFormat)}.",
        type=enum_type(OutputFormat),
    )

    parsed_args = parser.parse_args(args)

    scheme: type[VersioningScheme] = getattr(
        parsed_args, "versioning_scheme", PEP440VersioningScheme
    )

    if parsed_args.func in (create_release, show):
        # check for release-type
        if not getattr(parsed_args, "release_type", None):
            parser.error("--release-type is required.")

        if (
            getattr(parsed_args, "release_version", None)
            and parsed_args.release_type != ReleaseType.VERSION
        ):
            parser.error(
                "--release-version requires --release-type "
                f"{ReleaseType.VERSION.value}"
            )

        if parsed_args.release_type == ReleaseType.VERSION and not getattr(
            parsed_args, "release_version", None
        ):
            parser.error(
                f"--release-type {ReleaseType.VERSION.value} requires to set "
                "--release-version"
            )

        next_version = getattr(parsed_args, "next_version", None)
        if next_version:
            parsed_args.next_version = scheme.parse_version(next_version)

    release_version = getattr(parsed_args, "release_version", None)
    if release_version:
        parsed_args.release_version = scheme.parse_version(release_version)

    token = os.environ.get("GITHUB_TOKEN")
    user = os.environ.get("GITHUB_USER")
    return user, token, parsed_args



# -*- coding: utf-8 -*-
# pontos/release/release.py
# Copyright (C) 2020-2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import subprocess
import sys
from typing import NoReturn

from pontos.git import GitError
from pontos.release.parser import parse_args
from pontos.terminal.null import NullTerminal
from pontos.terminal.rich import RichTerminal


def main(
    args=None,
) -> NoReturn:
    username, token, parsed_args = parse_args(args)
    logging.basicConfig(format="%(levelname)s - %(name)s - %(message)s")

    if parsed_args.quiet:
        term = NullTerminal()
        error_terminal = NullTerminal()
        logging.disable()
    else:
        term = RichTerminal()  # type: ignore[assignment]
        error_terminal = RichTerminal(file=sys.stderr)  # type: ignore[assignment] # noqa: E501

    try:
        retval = parsed_args.func(
            parsed_args,
            terminal=term,
            error_terminal=error_terminal,
            username=username,
            token=token,
        )
        sys.exit(int(retval))
    except KeyboardInterrupt:
        sys.exit(1)
    except GitError as e:
        error_terminal.error(f'Could not run git command "{e.cmd}".')
        error = e.stderr if e.stderr else e.stdout
        error_terminal.print(f"Output was: {error}")
        sys.exit(1)
    except subprocess.CalledProcessError as e:
        if "--passphrase" not in e.cmd:
            error_terminal.error(f'Could not run command "{e.cmd}".')
        else:
            error_terminal.error("Headless signing failed.")

        error_terminal.print(f"Error was: {e.stderr}")
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from abc import ABC
from datetime import datetime, timezone
from types import TracebackType
from typing import Any, Dict, Optional, Type, Union

from httpx import AsyncClient, Response, Timeout

from pontos.helper import snake_case

SLEEP_TIMEOUT = 30.0  # in seconds
DEFAULT_TIMEOUT = 180.0  # three minutes
DEFAULT_TIMEOUT_CONFIG = Timeout(DEFAULT_TIMEOUT)  # three minutes

Headers = Dict[str, str]
Params = Dict[str, Union[str, int]]

__all__ = (
    "convert_camel_case",
    "format_date",
    "now",
    "NVDApi",
)


def now() -> datetime:
    """
    Return current datetime with UTC timezone applied
    """
    return datetime.now(tz=timezone.utc)


def format_date(date: datetime) -> str:
    """
    Format date matching to NVD api

    Args:
        date: Date to format

    Returns:
        Formatted date as string
    """
    return date.isoformat(timespec="seconds")


def convert_camel_case(dct: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert camel case keys into snake case keys

    Args:
        dct: dict to convert

    Returns:
        A dict with key names converted to snake case
    """
    converted = {}
    for key, value in dct.items():
        converted[snake_case(key)] = value
    return converted


async def sleep() -> None:
    await asyncio.sleep(SLEEP_TIMEOUT)


class NVDApi(ABC):
    """
    Abstract base class for querying the NIST NVD API.

    Should be used as an async context manager.
    """

    def __init__(
        self,
        url: str,
        *,
        token: Optional[str] = None,
        timeout: Optional[Timeout] = DEFAULT_TIMEOUT_CONFIG,
        rate_limit: bool = True,
    ) -> None:
        """
        Create a new instance of the CVE API.

        Args:
            url: The API URL to use.
            token: The API key to use. Using an API key allows to run more
                requests at the same time.
            timeout: Timeout settings for the HTTP requests
            rate_limit: Set to False to ignore rate limits. The public rate
                limit (without an API key) is 5 requests in a rolling 30 second
                window. The rate limit with an API key is 50 requests in a
                rolling 30 second window.
                See https://nvd.nist.gov/developers/start-here#divRateLimits
                Default: True.
        """
        self._url = url
        self._token = token
        self._client = AsyncClient(http2=True, timeout=timeout)

        if rate_limit:
            self._rate_limit: Optional[int] = 50 if token else 5
        else:
            self._rate_limit = None

        self._request_count = 0

    def _request_headers(self) -> Headers:
        """
        Get the default request headers
        """
        headers = {}

        if self._token:
            headers["apiKey"] = self._token

        return headers

    async def _consider_rate_limit(self) -> None:
        """
        Apply rate limit if necessary
        """
        if not self._rate_limit:
            return

        self._request_count += 1
        if self._request_count > self._rate_limit:
            await sleep()
            self._request_count = 0

    async def _get(
        self,
        *,
        params: Optional[Params] = None,
    ) -> Response:
        """
        A request against the NIST NVD CVE REST API.
        """
        headers = self._request_headers()

        await self._consider_rate_limit()

        return await self._client.get(self._url, headers=headers, params=params)

    async def __aenter__(self) -> "NVDApi":
        # reset rate limit counter
        self._request_count = 0
        await self._client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        return await self._client.__aexit__(  # type: ignore
            exc_type, exc_value, traceback
        )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Argument parser for pontos-github """

import os
from argparse import ArgumentParser, FileType, Namespace
from pathlib import Path
from typing import List, Optional

from pontos.github.cmds import (
    create_pull_request,
    create_release,
    create_tag,
    file_status,
    labels,
    pull_request,
    release,
    repos,
    tag,
    update_pull_request,
)
from pontos.github.models.base import FileStatus
from pontos.github.models.organization import RepositoryType

body_template = Path(__file__).parent / "pr_template.md"


def from_env(name: str) -> str:
    return os.environ.get(name, name)


def get_repository_type(rtype: str) -> RepositoryType:
    return RepositoryType[rtype]


def parse_args(
    args: Optional[List[str]] = None,
) -> Namespace:
    """
    Parsing args for Pontos GitHub

    Arguments:
    args        The program arguments passed by exec
    term        The terminal to print
    """

    parser = ArgumentParser(
        description="Greenbone GitHub API.",
    )

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Don't print messages to the terminal",
    )

    parser.add_argument(
        "--log-file",
        dest="log_file",
        type=str,
        help="Acivate logging using the given file path",
    )

    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        required=True,
        help="additional help",
        dest="command",
    )

    # create a PR from command line
    pr_parser = subparsers.add_parser(
        "pull-request", aliases=["pr", "PR", "pullrequest"]
    )

    pr_parser.set_defaults(func=pull_request)

    pr_parser.add_argument(
        "-t",
        "--token",
        default="GITHUB_TOKEN",
        type=from_env,
        help=(
            "GitHub Token to access the repository. "
            "Default looks for environment variable 'GITHUB_TOKEN'"
        ),
    )

    pr_subparsers = pr_parser.add_subparsers(
        title="method",
        dest="pr_method",
        metavar="name",
        description="valid pull request method",
        help="pull request method",
        required=True,
    )

    create_pr_parser = pr_subparsers.add_parser(
        "create", help="Create Pull Request"
    )

    create_pr_parser.set_defaults(pr_func=create_pull_request)

    create_pr_parser.add_argument(
        "repo", help="GitHub repository (owner/name) to use"
    )

    create_pr_parser.add_argument(
        "head",
        help=("Branch to create a pull request from"),
    )

    create_pr_parser.add_argument(
        "target",
        default="main",
        help="Branch as as target for the pull. Default: %(default)s",
    )

    create_pr_parser.add_argument(
        "title",
        help="Title for the pull request",
    )

    create_pr_parser.add_argument(
        "-b",
        "--body",
        default=body_template.read_text(encoding="utf-8"),
        help=(
            "Description for the pull request. Can be formatted in Markdown."
        ),
    )

    update_pr_parser = pr_subparsers.add_parser(
        "update", help="update Pull Request"
    )

    update_pr_parser.set_defaults(pr_func=update_pull_request)

    update_pr_parser.add_argument(
        "repo", help="GitHub repository (owner/name) to use"
    )
    update_pr_parser.add_argument(
        "pull_request", type=int, help="Pull Request to update"
    )
    update_pr_parser.add_argument(
        "--target",
        help="Branch as as target for the pull.",
    )
    update_pr_parser.add_argument(
        "--title",
        help="Title for the pull request",
    )

    update_pr_parser.add_argument(
        "-b",
        "--body",
        help=(
            "Description for the pull request. Can be formatted in Markdown."
        ),
    )

    # get files
    file_status_parser = subparsers.add_parser(
        "file-status", aliases=["status", "FS"]
    )

    file_status_parser.set_defaults(func=file_status)

    file_status_parser.add_argument(
        "repo", help=("GitHub repository (owner/name) to use")
    )

    file_status_parser.add_argument(
        "pull_request", help="Specify the Pull Request number", type=int
    )

    file_status_parser.add_argument(
        "-s",
        "--status",
        choices=FileStatus,
        default=[FileStatus.ADDED, FileStatus.MODIFIED],
        nargs="+",
        help="What file status should be returned. Default: %(default)s",
    )

    file_status_parser.add_argument(
        "-o",
        "--output",
        type=FileType("w", encoding="utf-8"),
        help=(
            "Specify an output file. "
            "If none is given, output will be prompted"
            "The file will contain all files, with status "
            "changes, as given, separated by a newline"
        ),
    )

    file_status_parser.add_argument(
        "-t",
        "--token",
        default="GITHUB_TOKEN",
        type=from_env,
        help=(
            "GitHub Token to access the repository. "
            "Default looks for environment variable 'GITHUB_TOKEN'"
        ),
    )

    # labels
    label_parser = subparsers.add_parser("labels", aliases=["L"])

    label_parser.set_defaults(func=labels)

    label_parser.add_argument(
        "repo", help="GitHub repository (owner/name) to use"
    )

    label_parser.add_argument(
        "issue", help="Specify the Issue/Pull Request number", type=int
    )

    label_parser.add_argument(
        "--labels",
        "-L",
        nargs="+",
        help="Specify the labels, that should be set",
    )

    label_parser.add_argument(
        "-t",
        "--token",
        default="GITHUB_TOKEN",
        type=from_env,
        help=(
            "GitHub Token to access the repository. "
            "Default looks for environment variable 'GITHUB_TOKEN'"
        ),
    )

    # orga-repos
    repos_parser = subparsers.add_parser("repos", aliases=["R"])

    repos_parser.set_defaults(func=repos)

    repos_parser.add_argument("orga", help="GitHub organization to use")

    repos_parser.add_argument(
        "-t",
        "--token",
        default="GITHUB_TOKEN",
        type=from_env,
        help=(
            "GitHub Token to access the repository. "
            "Default looks for environment variable 'GITHUB_TOKEN'"
        ),
    )

    repos_parser.add_argument(
        "--type",
        choices=RepositoryType,
        type=get_repository_type,
        default=RepositoryType.PUBLIC,
        help=(
            "Define the type of repositories that should be covered. "
            "Default: %(default)s"
        ),
    )

    repos_parser.add_argument(
        "-p",
        "--path",
        help="Define the Path to save the Repository Information",
    )

    # create a release from command line
    re_parser = subparsers.add_parser(
        "release", aliases=["re", "RE", "release"]
    )

    re_parser.set_defaults(func=release)

    re_parser.add_argument(
        "-t",
        "--token",
        default="GITHUB_TOKEN",
        type=from_env,
        help=(
            "GitHub Token to access the repository. "
            "Default looks for environment variable 'GITHUB_TOKEN'"
        ),
    )

    re_subparsers = re_parser.add_subparsers(
        title="method",
        dest="re_method",
        metavar="name",
        description="valid release method",
        help="Release method",
        required=True,
    )

    create_re_parser = re_subparsers.add_parser("create", help="Create release")

    create_re_parser.set_defaults(re_func=create_release)

    create_re_parser.add_argument(
        "repo", help="GitHub repository (owner/name) to use"
    )

    create_re_parser.add_argument(
        "tag",
        help="Tag to use for release",
    )

    create_re_parser.add_argument(
        "name",
        help="Name of the release",
    )

    create_re_parser.add_argument(
        "-b",
        "--body",
        default=None,
        help="Description for the Release. Can be formatted in Markdown.",
    )

    create_re_parser.add_argument(
        "-tc",
        "--target-commitish",
        default=None,
        help="Git reference to use for the release",
    )

    create_re_parser.add_argument(
        "-d",
        "--draft",
        action="store_true",
        default=False,
        help="Create a draft release.",
    )

    create_re_parser.add_argument(
        "-p",
        "--prerelease",
        action="store_true",
        default=False,
        help="Create a pre-release.",
    )

    # Create a tag from command line
    tag_parser = subparsers.add_parser("tag", aliases=["tag", "TAG"])

    tag_parser.set_defaults(func=tag)

    tag_parser.add_argument(
        "-t",
        "--token",
        default="GITHUB_TOKEN",
        type=from_env,
        help=(
            "GitHub Token to access the repository. "
            "Default looks for environment variable 'GITHUB_TOKEN'"
        ),
    )

    tag_subparsers = tag_parser.add_subparsers(
        title="method",
        dest="tag_method",
        metavar="name",
        description="valid tag method",
        help="Release method",
        required=True,
    )

    create_tag_parser = tag_subparsers.add_parser("create", help="Create tag")

    create_tag_parser.set_defaults(tag_func=create_tag)

    create_tag_parser.add_argument(
        "repo", help="GitHub repository (owner/name) to use"
    )

    create_tag_parser.add_argument(
        "tag",
        help="Tag name to use",
    )

    create_tag_parser.add_argument(
        "name",
        help="Name of the user",
    )

    create_tag_parser.add_argument(
        "message",
        help="Tag message",
    )

    create_tag_parser.add_argument(
        "git_object",
        help="The SHA of the git object this is tagging.",
    )
    create_tag_parser.add_argument(
        "email",
        help="Email address of the user",
    )

    create_tag_parser.add_argument(
        "-got",
        "--git-object-type",
        default="commit",
        help="The type of the object we're taggin",
    )

    create_tag_parser.add_argument(
        "-d",
        "--date",
        default=None,
        help=(
            "When this object was tagged. ISO 8601 format:"
            " YYYY-MM-DDTHH:MM:SSZ."
        ),
    )

    parsed_args = parser.parse_args(args)

    return parsed_args



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from argparse import ArgumentParser, Namespace
from typing import Callable

from pontos.nvd.cpe.api import CPEApi

__all__ = ("CPEApi",)


async def query_cpe(args: Namespace) -> None:
    async with CPEApi(token=args.token) as api:
        cpe = await api.cpe(args.cpe_name_id)
        print(cpe)


async def query_cpes(args: Namespace) -> None:
    async with CPEApi(token=args.token) as api:
        async for cpe in api.cpes(
            keywords=args.keywords, cpe_match_string=args.cpe_match_string
        ):
            print(cpe)


def cpe_main() -> None:
    parser = ArgumentParser()
    parser.add_argument("--token", help="API key to use for querying.")
    parser.add_argument(
        "cpe_name_id", metavar="CPE Name ID", help="UUID of the CPE"
    )

    main(parser, query_cpe)


def cpes_main() -> None:
    parser = ArgumentParser()
    parser.add_argument("--token", help="API key to use for querying.")
    parser.add_argument(
        "--cpe-match-string",
        help="Search for CPE names that exist in the Official CPE Dictionary.",
    )
    parser.add_argument(
        "--keywords",
        nargs="*",
        help="Search for CPEs containing the keyword in their titles and "
        "references.",
    )

    main(parser, query_cpes)


def main(parser: ArgumentParser, func: Callable) -> None:
    try:
        args = parser.parse_args()
        asyncio.run(func(args))
    except KeyboardInterrupt:
        pass



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This script downloads a single artifacts of a given repository
"""

from argparse import ArgumentParser, Namespace
from pathlib import Path

from rich.progress import Progress

from pontos.github.api import GitHubAsyncRESTApi


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument("repository")
    parser.add_argument("tag", help="Release Tag")
    parser.add_argument(
        "--file",
        help="File to write the artifact to. Default: %(default)s",
        default="out.file",
        type=Path,
    )
    parser.add_argument(
        "--type",
        choices=["zip", "tar"],
        help="Download release asset type. Default: %(default)s",
        default="tar",
    )


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    with args.file.open("wb") as f, Progress() as rich_progress:
        task_id = rich_progress.add_task(
            f"[red]Downloading asset for tag {args.tag} as {args.type}... ",
            total=None,
        )
        download_api = (
            api.releases.download_release_tarball
            if args.type == "tar"
            else api.releases.download_release_zip
        )
        async with download_api(args.repository, args.tag) as download:
            async for content, progress in download:
                rich_progress.advance(task_id, progress or 1)
                f.write(content)

            rich_progress.update(task_id, total=1, completed=1)

    return 0



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import AsyncIterator, Iterable, Union

from pontos.github.api.client import GitHubAsyncREST
from pontos.github.models.organization import Repository
from pontos.github.models.search import Qualifier, RepositorySort, SortOrder
from pontos.helper import enum_or_value


class GitHubAsyncRESTSearch(GitHubAsyncREST):
    async def repositories(
        self,
        *,
        keywords: Iterable[str],
        qualifiers: Iterable[Qualifier],
        order: Union[str, SortOrder] = SortOrder.DESC,
        sort: Union[str, RepositorySort, None] = None,
    ) -> AsyncIterator[Repository]:
        """
        Search for repositories

        https://docs.github.com/en/rest/search#search-repositories

        https://docs.github.com/en/search-github/searching-on-github/searching-for-repositories

        Args:
            keywords: List of keywords to search for.
            qualifiers: List of qualifiers.
            order: Sort order either 'asc' or 'desc'. Default is 'desc'.
            sort: Sort the found repositories by this criteria.

        Raises:
            `httpx.HTTPStatusError`: If there was an error in the request

        Returns:
            An async iterator yielding repositories

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi
                from pontos.github.models import (
                    InReadmeQualifier,
                    InTopicsQualifier,
                )

                async with GitHubAsyncRESTApi(token) as api:
                    # search for keywords in repo topics and READMEs
                    async for repo in api.search(
                        keywords=["utils", "search", "golang"],
                        qualifier=[
                            InTopicsQualifier(),
                            InReadmeQualifier(),
                        ],
                    )
        """
        api = "/search/repositories"
        params = {
            "per_page": "100",
        }
        if order:
            params["order"] = enum_or_value(order)
        if sort:
            params["sort"] = enum_or_value(sort)

        query = (
            f"{' '.join(keywords)} "
            f"{' '.join([str(qualifier) for qualifier in qualifiers])}"
        )
        params["q"] = query

        async for response in self._client.get_all(api, params=params):
            response.raise_for_status()
            data = response.json()
            for repo in data["items"]:
                yield Repository.from_dict(repo)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import IO, Any, Callable, Iterable, Optional

from rich.console import Console, RenderableType
from rich.padding import Padding
from rich.progress import (
    BarColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskProgressColumn,
    TextColumn,
)

from pontos.helper import DownloadProgressIterable

from .terminal import Signs, Terminal


def red(text: str) -> str:
    return f"[red]{text}[/red]"


def yellow(text: str) -> str:
    return f"[yellow]{text}[/yellow]"


def cyan(text: str) -> str:
    return f"[cyan]{text}[/cyan]"


def green(text: str) -> str:
    return f"[green]{text}[/green]"


def white(text: str) -> str:
    return f"[white]{text}[/white]"


class PaddingColumn(ProgressColumn):
    def __init__(self, indent: int, table_column=None):
        self._padding = Padding.indent("", indent)
        super().__init__(table_column=table_column)

    def render(self, task: Task) -> RenderableType:
        return self._padding


class RichTerminal(Terminal):
    """
    A Terminal based on `rich <https://github.com/Textualize/rich/>`_.
    """

    def __init__(
        self,
        file: Optional[IO[str]] = None,
    ) -> None:
        """
        Create a new RichTerminal

        Args:
            file: A file object where the output should write to.
                Default is stdout.
        """
        super().__init__()
        self._console = Console(file=file)

    def _indent_message(self):
        return " " * self._indent

    def _print_status(
        self,
        *messages: Any,
        status: Signs,
        color: Callable,
        **kwargs: Any,
    ):
        self._console.print(
            self._indent_message(), color(status), *messages, **kwargs
        )

    def get_progress_default_columns(self) -> Iterable[ProgressColumn]:
        return (
            PaddingColumn(self._indent),
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
        )

    def progress(
        self,
        *,
        columns: Optional[Iterable[ProgressColumn]] = None,
        additional_columns: Optional[Iterable[ProgressColumn]] = None,
        **kwargs,
    ) -> Progress:
        kwargs["console"] = self._console
        columns = columns or self.get_progress_default_columns()
        if additional_columns:
            columns = *columns, *additional_columns
        return Progress(
            *columns,
            **kwargs,
        )

    def out(self, *messages: Any, **kwargs: Any) -> None:
        kwargs["highlight"] = False
        self._console.out(self._indent_message(), *messages, **kwargs)

    def print(self, *messages: Any, **kwargs: Any) -> None:
        self._console.print(self._indent_message(), *messages, **kwargs)

    def ok(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.OK, "color": green})
        self._print_status(*messages, **kwargs)

    def fail(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.FAIL, "color": red})
        self._print_status(*messages, **kwargs)

    def error(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.ERROR, "color": red})
        self._print_status(*messages, **kwargs)

    def warning(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.WARNING, "color": yellow})
        self._print_status(*messages, **kwargs)

    def info(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.INFO, "color": cyan})
        self._print_status(*messages, **kwargs)

    def bold_info(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.INFO, "color": cyan, "style": "bold"})
        self._print_status(*messages, **kwargs)

    def download_progress(self, progress: DownloadProgressIterable) -> None:
        with self.progress() as rich_progress:
            task_description = f"Downloading [blue]{progress.url}"
            if progress.length:
                task_id = rich_progress.add_task(
                    task_description, total=progress.length
                )
                for percent in progress:
                    rich_progress.advance(task_id, percent)  # type: ignore[arg-type] # noqa: E501
            else:
                task_id = rich_progress.add_task(task_description, total=None)
                for _ in progress:
                    rich_progress.advance(task_id)

            rich_progress.update(task_id, total=1, completed=1)



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from typing import List, Literal, Union

from .commands import VersionCommand, get_commands
from .errors import ProjectError
from .schemes import VersioningScheme
from .version import Version, VersionUpdate

__all__ = ("Project",)


class Project:
    """
    A project for handling versioning

    Example:
        .. code-block:: python

            from pontos.version.scheme import PEP440VersioningScheme
            from pontos.version.project import Project

            project = Project(PEP440VersioningScheme)
    """

    def __init__(self, versioning_scheme: VersioningScheme) -> None:
        """
        Creates a new project instance

        Args:
            versioning_scheme: Scheme for version handling

        Raises:
            ProjectError: If no fitting VersionCommand could be found
        """
        self._versioning_scheme = versioning_scheme
        self._commands = self._gather_commands()

    def _gather_commands(self) -> List[VersionCommand]:
        """
        Initialize the project with the fitting VersionCommands of the current
        working directory

        Raises:
            ProjectError: If no fitting VersionCommand could be found
        """
        commands = []
        for cmd in get_commands():
            command = cmd(versioning_scheme=self._versioning_scheme)
            if command.project_found():
                commands.append(command)

        if not commands:
            raise ProjectError("No project settings file found")

        return commands

    def update_version(
        self, new_version: Version, *, force: bool = False
    ) -> VersionUpdate:
        """
        Update the current version of this project

        Args:
            new_version: Use this version in the update
            force: Force updating the version even if the current version is the
                same as the new version

        Returns:
            The version update including the changed files
        """
        update = self._commands[0].update_version(new_version, force=force)
        for cmd in self._commands[1:]:
            next_update = cmd.update_version(new_version, force=force)
            update.changed_files.extend(next_update.changed_files)

        return update

    def get_current_version(self) -> Version:
        """
        Get the current version of the project

        Returns:
            The current version
        """
        return self._commands[0].get_current_version()

    def verify_version(
        self, version: Union[Literal["current"], Version]
    ) -> None:
        """
        Verify the current version of this project

        Args:
            version: Version to check against the current applied version of
                this project. If version is "current" the command should verify
                if all version information is consistent, for example if the
                version information in several files is the same.
        """
        for cmd in self._commands:
            cmd.verify_version(version)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=line-too-long

from typing import Any, Dict, Optional


def get_cve_data(data: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    cve = {
        "id": "CVE-2022-45536",
        "source_identifier": "cve@mitre.org",
        "published": "2022-11-22T21:15:11.103",
        "last_modified": "2022-11-23T16:02:07.367",
        "vuln_status": "Analyzed",
        "descriptions": [
            {
                "lang": "en",
                "value": "AeroCMS v0.0.1 was discovered to contain a SQL "
                "Injection vulnerability via the id parameter at "
                "\\admin\\post_comments.php. This vulnerability allows "
                "attackers to access database information.",
            }
        ],
        "references": [
            {
                "url": "https://github.com/rdyx0/CVE/blob/master/AeroCMS/AeroCMS-v0.0.1-SQLi/post_comments_sql_injection/post_comments_sql_injection.md",
                "source": "cve@mitre.org",
                "tags": ["Exploit", "Third Party Advisory"],
            },
            {
                "url": "https://rdyx0.github.io/2018/09/07/AeroCMS-v0.0.1-SQLi%20post_comments_sql_injection/",
                "source": "cve@mitre.org",
                "tags": ["Exploit", "Third Party Advisory"],
            },
        ],
    }
    if data:
        cve.update(data)
    return cve


def get_cpe_data(update: Dict[str, Any] = None) -> Dict[str, Any]:
    data = {
        "deprecated": False,
        "cpe_name": "cpe:2.3:o:microsoft:windows_10_22h2:-:*:*:*:*:*:arm64:*",
        "cpe_name_id": "9BAECDB2-614D-4E9C-9936-190C30246F03",
        "last_modified": "2022-12-09T18:15:16.973",
        "created": "2022-12-09T16:20:06.943",
    }

    if update:
        data.update(update)
    return data



# Copyright (C) 2020-2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import asyncio
import hashlib
import subprocess
from argparse import Namespace
from asyncio.subprocess import Process
from enum import IntEnum
from os import PathLike
from pathlib import Path
from typing import AsyncContextManager, Optional, SupportsInt, Union

import httpx
from rich.progress import Progress as RichProgress
from rich.progress import TextColumn

from pontos.errors import PontosError
from pontos.git.git import GitError
from pontos.github.api import GitHubAsyncRESTApi
from pontos.helper import AsyncDownloadProgressIterable
from pontos.release.command import AsyncCommand
from pontos.terminal import Terminal
from pontos.version import Version
from pontos.version.helper import get_last_release_version
from pontos.version.schemes import VersioningScheme

from .helper import get_git_repository_name


class SignReturnValue(IntEnum):
    """
    Possible return values of SignCommand
    """

    SUCCESS = 0
    TOKEN_MISSING = 1
    NO_PROJECT = 2
    NO_RELEASE_VERSION = 3
    NO_RELEASE = 4
    UPLOAD_ASSET_ERROR = 5
    SIGNATURE_GENERATION_FAILED = 6


class SignatureError(PontosError):
    """
    Error while creating a signature
    """


async def cmd_runner(*args: Union[str, PathLike[str]]) -> Process:
    return await asyncio.create_subprocess_exec(
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )


class SignCommand(AsyncCommand):
    """
    A CLI command for signing a release

    Args:
        terminal: A Terminal for output
    """

    async def _async_download_progress(
        self,
        rich_progress: RichProgress,
        progress: AsyncDownloadProgressIterable[bytes],
        destination: Path,
    ) -> None:
        with destination.open("wb") as f:
            task_description = f"Downloading [blue]{progress.url}"
            task_id = rich_progress.add_task(
                task_description,
                total=progress.length,
                sha256="",
            )
            sha256 = hashlib.sha256()
            async for content, percent in progress:
                rich_progress.advance(task_id, percent or 1)
                f.write(content)
                sha256.update(content)

            rich_progress.update(
                task_id, total=1, completed=1, sha256=sha256.hexdigest()
            )

    async def download_zip(
        self,
        rich_progress: RichProgress,
        github: GitHubAsyncRESTApi,
        destination: Path,
        repo: str,
        git_version: str,
    ) -> Path:
        async with github.releases.download_release_zip(
            repo, git_version
        ) as download:
            await self._async_download_progress(
                rich_progress, download, destination
            )
        return destination

    async def download_tar(
        self,
        rich_progress: RichProgress,
        github: GitHubAsyncRESTApi,
        destination: Path,
        repo: str,
        git_version: str,
    ) -> Path:
        async with github.releases.download_release_tarball(
            repo, git_version
        ) as download:
            await self._async_download_progress(
                rich_progress, download, destination
            )
        return destination

    async def download_asset(
        self,
        rich_progress: RichProgress,
        name: str,
        download_cm: AsyncContextManager[AsyncDownloadProgressIterable[bytes]],
    ) -> Path:
        file_path = Path(name)
        async with download_cm as iterator:
            await self._async_download_progress(
                rich_progress, iterator, file_path
            )
        return file_path

    async def sign_file(
        self, file_path: Path, signing_key: str, passphrase: Optional[str]
    ) -> None:
        self.terminal.info(f"Signing {file_path}")

        if passphrase:
            process = await cmd_runner(
                "gpg",
                "--pinentry-mode",
                "loopback",
                "--default-key",
                signing_key,
                "--yes",
                "--detach-sign",
                "--passphrase",
                passphrase,
                "--armor",
                file_path,
            )
        else:
            process = await cmd_runner(
                "gpg",
                "--default-key",
                signing_key,
                "--yes",
                "--detach-sign",
                "--armor",
                file_path,
            )

        _, stderr = await process.communicate()
        if process.returncode:
            raise SignatureError(
                f"Could not create signature for {file_path}. "
                f"{stderr.decode(errors='replace')}"
            )

    async def async_run(  # type: ignore[override]
        self,
        *,
        token: str,
        space: str,
        versioning_scheme: VersioningScheme,
        signing_key: str,
        passphrase: str,
        dry_run: Optional[bool] = False,
        project: Optional[str],
        git_tag_prefix: Optional[str],
        release_version: Optional[Version],
        release_series: Optional[str] = None,
    ) -> SignReturnValue:
        """
        Sign a release

        Args:
            token: A token for creating a release on GitHub
            dry_run: True to not upload the signature files
            space: GitHub username or organization. Required for generating
                links in the changelog.
            project: Name of the project to release. If not set it will be
                gathered via the git remote url.
            versioning_scheme: The versioning scheme to use for version parsing
                and calculation
            git_tag_prefix: An optional prefix to use for handling a git tag
                from the release version.
            release_version: Optional release version to use. If not set the
                current version will be determined from the project.
            signing_key: A GPG key ID to use for creating signatures.
            passphrase: Passphrase for the signing key
            release_series: Optional release series to use.
                For example: "1.2", "2", "23".
        """
        if not token and not dry_run:
            # dry run doesn't upload assets. therefore a token MAY NOT be
            # required # for public repositories.
            self.print_error(
                "Token is missing. The GitHub token is required to upload "
                "signature files."
            )
            return SignReturnValue.TOKEN_MISSING

        self.terminal.info(f"Using versioning scheme {versioning_scheme.name}")

        try:
            project = (
                project if project is not None else get_git_repository_name()
            )
        except GitError as e:
            self.print_error(f"Could not determine project. {e}")
            return SignReturnValue.NO_PROJECT

        try:
            release_version = (
                release_version
                if release_version is not None
                else get_last_release_version(
                    versioning_scheme.parse_version,
                    git_tag_prefix=git_tag_prefix,
                    tag_name=f"{git_tag_prefix}{release_series}.*"
                    if release_series
                    else None,
                )
            )
        except PontosError as e:
            self.print_error(f"Could not determine release version. {e}")
            return SignReturnValue.NO_RELEASE_VERSION

        if not release_version:
            return SignReturnValue.NO_RELEASE_VERSION

        git_version: str = f"{git_tag_prefix}{release_version}"
        repo = f"{space}/{project}"

        async with GitHubAsyncRESTApi(token=token) as github:
            if not await github.releases.exists(repo, git_version):
                self.print_error(
                    f"Release version {git_version} does not exist."
                )
                return SignReturnValue.NO_RELEASE

            tasks = []

            zip_destination = Path(f"{project}-{release_version}.zip")
            tarball_destination = Path(f"{project}-{release_version}.tar.gz")

            # terminal can be a NullTerminal here too that doesn't have a
            # progress. this needs to be fixed and the type ignore removed
            # afterwards
            with self.terminal.progress(  # type: ignore[attr-defined]
                additional_columns=[
                    TextColumn("[progress.description]{task.fields[sha256]}"),
                ]
            ) as rich_progress:
                tasks.append(
                    asyncio.create_task(
                        self.download_zip(
                            rich_progress,
                            github,
                            zip_destination,
                            repo,
                            git_version,
                        )
                    )
                )
                tasks.append(
                    asyncio.create_task(
                        self.download_tar(
                            rich_progress,
                            github,
                            tarball_destination,
                            repo,
                            git_version,
                        )
                    )
                )

                # pylint: disable=line-too-long
                async for name, download_cm in github.releases.download_release_assets(  # noqa: E501
                    repo,
                    git_version,
                ):
                    tasks.append(
                        asyncio.create_task(
                            self.download_asset(
                                rich_progress, name, download_cm
                            )
                        )
                    )

                file_paths = await asyncio.gather(*tasks)

            tasks = [
                asyncio.create_task(
                    self.sign_file(file_path, signing_key, passphrase)
                )
                for file_path in file_paths
            ]

            done, pending = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_EXCEPTION
            )

            has_error = False
            for task in done:
                try:
                    await task
                except (asyncio.CancelledError, asyncio.InvalidStateError):
                    pass
                except SignatureError as e:
                    self.print_error(e)
                    has_error = True

            for task in pending:
                # we had an error
                try:
                    task.cancel()
                    await task
                except asyncio.CancelledError:
                    pass

            if has_error:
                return SignReturnValue.SIGNATURE_GENERATION_FAILED

            if dry_run:
                return SignReturnValue.SUCCESS

            upload_files = [
                (Path(f"{str(p)}.asc"), "application/pgp-signature")
                for p in file_paths
            ]
            self.terminal.info(
                f"Uploading assets: {[str(p[0]) for p in upload_files]}"
            )

            try:
                # pylint: disable=line-too-long
                async for uploaded_file in github.releases.upload_release_assets(  # noqa: E501
                    repo, git_version, upload_files
                ):
                    self.terminal.ok(f"Uploaded: {uploaded_file}")
            except httpx.HTTPStatusError as e:
                self.print_error(f"Failed uploading asset {e}.")
                return SignReturnValue.UPLOAD_ASSET_ERROR

        return SignReturnValue.SUCCESS


def sign(
    args: Namespace,
    *,
    terminal: Terminal,
    error_terminal: Terminal,
    token: str,
    **_kwargs,
) -> SupportsInt:
    return SignCommand(terminal=terminal, error_terminal=error_terminal).run(
        token=token,
        dry_run=args.dry_run,
        project=args.project,
        space=args.space,
        versioning_scheme=args.versioning_scheme,
        git_tag_prefix=args.git_tag_prefix,
        release_version=args.release_version,
        signing_key=args.signing_key,
        passphrase=args.passphrase,
        release_series=args.release_series,
    )



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from pontos.typing import SupportsStr


class SupportsStrTestCase(unittest.TestCase):
    def test_str(self):
        self.assertIsInstance("", SupportsStr)
        self.assertIsInstance(None, SupportsStr)

    def test_some_class(self):
        class Foo:
            def __str__(self) -> str:
                pass

        foo = Foo()

        self.assertIsInstance(foo, SupportsStr)



# Copyright (C) 2021-2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import re
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Union

from ..errors import VersionError
from ..version import Version, VersionUpdate
from ._command import VersionCommand


# This class is used for JavaScript Version command(s)
class JavaScriptVersionCommand(VersionCommand):
    project_file_name = "package.json"
    version_file_paths = (Path("src", "version.js"), Path("src", "version.ts"))
    _package = None

    @property
    def package(self) -> Dict[str, Any]:
        if self._package:
            return self._package

        if not self.project_file_path.exists():
            raise VersionError(f"{self.project_file_path} file not found.")

        try:
            with self.project_file_path.open(mode="r", encoding="utf-8") as fp:
                self._package = json.load(fp)
        except OSError as e:
            raise VersionError(
                "No version tag found. Maybe this "
                "module has not been released at all."
            ) from e
        except json.JSONDecodeError as e:
            raise VersionError(
                "No valid JSON found. Maybe this "
                "module has not been released at all."
            ) from e

        if not self._package.get("version", None):
            raise VersionError(
                f"Version field missing in {self.project_file_path}."
            )

        return self._package

    def _get_current_file_version(
        self, version_file: Path
    ) -> Optional[Version]:
        if not version_file.exists():
            return None

        content = version_file.read_text(encoding="utf-8")
        match = re.search(r'VERSION = "(?P<version>.*)";', content)
        if not match:
            raise VersionError(f"VERSION variable not found in {version_file}")

        return self.versioning_scheme.parse_version(match.group("version"))

    def get_current_version(self) -> Version:
        """Get the current version of this project
        In go the version is only defined within the repository
        tags, thus we need to check git, what tag is the latest"""
        return self.versioning_scheme.parse_version(self.package["version"])

    def verify_version(
        self, version: Union[Literal["current"], Version, None]
    ) -> None:
        """Verify the current version of this project"""
        current_version = self.get_current_version()

        if not version or version == "current":
            for version_file in self.version_file_paths:
                file_version = self._get_current_file_version(version_file)
                if file_version and file_version != current_version:
                    raise VersionError(
                        f"The version {file_version} in "
                        f"{version_file} doesn't match the current "
                        f"version {current_version}."
                    )
            return

        if current_version != version:
            raise VersionError(
                f"Provided version {version} does not match the "
                f"current version {current_version} in "
                f"{self.project_file_path}."
            )

        for version_file in self.version_file_paths:
            file_version = self._get_current_file_version(version_file)
            if file_version and file_version != version:
                raise VersionError(
                    f"Provided version {version} does not match the "
                    f"current version {file_version} in {version_file}."
                )

    def _update_package_json(self, new_version: Version) -> None:
        """
        Update the version in the package.json file
        """
        try:
            self.package["version"] = str(new_version)

            with self.project_file_path.open(mode="w") as fp:
                json.dump(obj=self.package, fp=fp, indent=2)

        except EnvironmentError as e:
            raise VersionError(
                "No version tag found. Maybe this "
                "module has not been released at all."
            ) from e
        except json.JSONDecodeError as e:
            raise VersionError("Couldn't load JSON") from e

    def _update_version_file(
        self, version_file: Path, new_version: Version
    ) -> bool:
        """
        Update the version file with the new version
        """
        if not version_file.exists():
            return False

        content = version_file.read_text(encoding="utf-8")
        match = re.search(
            pattern=r'VERSION = (?P<quote>[\'"])(?P<version>.*)(?P=quote)',
            string=content,
        )
        if not match:
            return False

        content = content.replace(match.group("version"), str(new_version))
        version_file.write_text(content, encoding="utf-8")
        return True

    def update_version(
        self, new_version: Version, *, force: bool = False
    ) -> VersionUpdate:
        try:
            package_version = self.get_current_version()
            if not force and new_version == package_version:
                return VersionUpdate(previous=package_version, new=new_version)
        except VersionError:
            # just ignore current version and override it
            package_version = None

        changed_files = [self.project_file_path]
        self._update_package_json(new_version=new_version)

        for version_file in self.version_file_paths:
            updated = self._update_version_file(
                version_file, new_version=new_version
            )
            if updated:
                changed_files.append(version_file)

        return VersionUpdate(
            previous=package_version,
            new=new_version,
            changed_files=changed_files,
        )
